
{*******************************************************}
{                                                       }
{       Animated Menus                                  }
{       TCustomPopupMenu2000 and TPopupMenu200Form      }
{                                                       }
{       Copyright (c) 1997-2001 AnimatedMenus.com       }
{       All rights reserved.                            }
{                                                       }
{*******************************************************}


unit am2000popupmenu;

{$I am2000.inc}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, StdCtrls, ExtCtrls,
  Forms, Dialogs, Buttons, ComCtrls, Menus,
  am2000options, am2000menuitem, am2000utils;


type
  // popup menu
  TCustomPopupMenu2000 = class(T_AM2000_PopupMenu)
  private
    FEMI2000         : TEditableMenuItem2000; // layer for comptibility with Delphi and your apps
    FSBPanelNo       : Integer;
    FStatusBar       : TStatusBar;
    FOnMenuCommand   : TNotifyEvent;
    FOnMenuClose     : T_AM2000_MenuCloseEvent;
    FFont            : TFont;
    FParentFont      : Boolean;
    FParentShowHint  : Boolean;
    FShowHint        : Boolean;
    FSystemFont      : Boolean;
    FOnCloseQuery    : TCloseQueryEvent;
    FOptions         : T_AM2000_Options;
    FCtl3D           : Boolean;
    PopupX           : Integer;
    PopupY           : Integer;
    FComponentItem          : TMenuItem2000;
    FComponentItems         : TList;
    FComponentItemsParent   : TMenuItem;
    FComponentItemsMenuIndex: Integer;
    FHideProperty    : Integer;
    FOnContextMenu   : TNotifyEvent;
    FForm            : TForm;

{$IFNDEF Delphi4OrHigher}
    FImages          : TImageList;
{$ENDIF}

    function IsFontStored: Boolean;
    function IsShowHintStored: Boolean;
    function GetMenuItem2000: TMenuItem2000;
    function GetForm: TForm;
    procedure SetFont(Value: TFont);
    procedure SetParentFont(Value: Boolean);
    procedure SetShowHint(Value: Boolean);
    procedure SetParentShowHint(Value: Boolean);
    procedure SetSystemFont(Value: Boolean);
    procedure SetOptions(Value: T_AM2000_Options);
    procedure SetVersion(Value: T_AM2000_Version);
    function GetVersion: T_AM2000_Version;

  protected
    procedure AssignMenuItems(var DestItems: TMenuItem; var DestHandle: HMenu); virtual;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Loaded; override;

    procedure DoCloseQuery(Sender: TObject; var CanClose: Boolean); dynamic;

    function GetComponentItemsCaption: String; virtual;
    procedure CreateComponentItems(Items: TMenuItem2000; AddEmpty: Boolean); virtual;
    procedure UpdateComponentItems(Items: TMenuItem2000); virtual;

    property Version: T_AM2000_Version
      read GetVersion write SetVersion stored False;

  public
    MenuItems        : TMenuItem;
    MenuHandle       : HMenu;
    Ole2ObjectWindow : HWnd;
    MenuPopulated    : Boolean;
    IgnoreAnimation, HotTrack, ShowAccelerators, KeepDesktop: Boolean;

    property StatusBar       : TStatusBar
      read FStatusBar write FStatusBar;
    property StatusBarIndex  : Integer
      read FSBPanelNo write FSBPanelNo default 0;
    property ShowHint        : Boolean
      read FShowHint write SetShowHint stored IsShowHintStored;
    property Options         : T_AM2000_Options
      read FOptions write SetOptions;
    property Font            : TFont
      read FFont write SetFont stored IsFontStored;
    property SystemFont      : Boolean
      read FSystemFont write SetSystemFont default True;
    property Ctl3D           : Boolean
      read FCtl3D write FCtl3D default True;
    property ParentShowHint  : Boolean
      read FParentShowHint write SetParentShowHint default True;
    property ParentFont      : Boolean
      read FParentFont write SetParentFont default False;

    property Items2000       : TMenuItem2000
      read GetMenuItem2000;

    property OnMenuCommand   : TNotifyEvent
      read FOnMenuCommand write FOnMenuCommand;
    property OnMenuClose     : T_AM2000_MenuCloseEvent
      read FOnMenuClose write FOnMenuClose;
    property OnCloseQuery    : TCloseQueryEvent
      read FOnCloseQuery write FOnCloseQuery;
    property OnContextMenu   : TNotifyEvent
      read FOnContextMenu write FOnContextMenu;

{$IFNDEF Delphi4OrHigher}
    property Images : TImageList read FImages write FImages;
{$ENDIF}

    property Form: TForm
      read GetForm;

    constructor Create(AOwner: TComponent); override;
    destructor Destroy;                     override;

    procedure Popup(X, Y: Integer);         override;
    procedure PopupAsync(X, Y: Integer);    virtual;
    procedure SetSelectedIndex(First: Boolean);
    procedure RemoveShowHiddenFlag;
    function FormOnScreen: Boolean;
    function GetTopMostForm: TForm;
    procedure PrepareForPopup(AddEmpty: Boolean); virtual;

    function IsShortCut(var Msg: TWMKey): Boolean;
{$IFDEF Delphi4OrHigher}
    override;
{$ENDIF}

    procedure Refresh;
    
  published
    property Items   : TEditableMenuItem2000  read FEMI2000;

    // hide these properties from ObjectInspector
    property Alignment: Integer read FHideProperty;
    property AutoMerge: Integer read FHideProperty;
    property AutoHotKeys: Integer read FHideProperty;
    property AutoLineReduction: Integer read FHideProperty;
    property AutoPopup: Integer read FHideProperty;
    property BiDiMode: Integer read FHideProperty;
    property HelpContext: Integer read FHideProperty;
    property MenuAnimation: Integer read FHideProperty;
    property OwnerDraw: Integer read FHideProperty;
    property ParentBiDiMode: Integer read FHideProperty;
    property Tag: Integer read FHideProperty;
    property TrackButton: Integer read FHideProperty;

  end;


  // template
  TCustomPopupMenu2000Form = class(TForm)
  end;

function GetPopupMenuForm(Item: TMenuItem): TCustomPopupMenu2000Form;

implementation
uses
  CommCtrl, MmSystem,
  am2000menubar, am2000title, am2000hintwindow, am2000cache, am2000dropshadow;

const
  COLOR_GRADIENTACTIVECAPTION = 27;
  COLOR_GRADIENTINACTIVECAPTION = 28;

  clGradientActiveCaption = TColor(COLOR_GRADIENTACTIVECAPTION or $80000000);
  clGradientInactiveCaption = TColor(COLOR_GRADIENTINACTIVECAPTION or $80000000);

type
  T_AM2000_PMFState = set of (fsSelectedChanged, fsMouseChanged, fsPaintMenu, fsFromBottomToTop,
    fsFromRightToLeft, fsDrawDisabled, fsIgnoreMouseMove, fsBecomingDraggable, fsAnimated,
    fsKillAnimate, fsDisabled, fsCtl3d, fsShowHidden, fsHiddenArrow, fsHiddenAsRegular,
    fsNoDrawCanvas, fsScrollBars, fsFraming, fsIgnoreNextMouseUp, fsIgnoreNextMenuUp, fsDesigning,
    fsShowAccelerators, fsClicked);

  TByteArray = array [0..1024*1024] of Byte;
  PByteArray = ^TByteArray;

  TShortIntArray = array [0..1024*1024] of ShortInt;
  PShortIntArray = ^TShortIntArray;

  TPopupMenu2000Form = class(TCustomPopupMenu2000Form)
  private
    bi: TBitmapInfo;
    bits0, bits1: PByteArray;
    bits, dbits: PShortIntArray;
    Buffer, Back: TBitmap;
    FSelectedIndex, LastSelectedIndex, ItemWidth, ShortcutWidth,
      ItemHeight, DX, DY, ParentMenuIndex, CurHiddenCount, BitsSize, CurStep,
      FirstVisibleIndex, LastVisibleIndex: Integer;
    Animation: T_AM2000_Animation;

    PopupMenu: TCustomPopupMenu2000;
    SubMenuForm, ParentMenuForm: TPopupMenu2000Form;
    MenuHandle: HMenu;
    Options: T_AM2000_Options;
    MouseState: T_AM2000_MouseState;

    MX, MY: Integer;
    TimeStart: Integer;
    State: T_AM2000_PMFState;
    ToolTipWindow: T_AM2000_ToolTipWindow;
    DesktopDC: HDC;

    // items rects
    FiLeft, FiWidth: Integer;
    ItemRects: TList;
    FocusItem: TMenuItem;

    DropShadow: T_AM2000_DropShadowForm;

    procedure wmKeyDown(var Msg: TWMKeyDown);      message wm_KeyDown;
    procedure wmSysKeyDown(var Msg: TWMKeyDown);   message wm_SysKeyDown;
    procedure wmChar(var Msg: TWMChar);            message wm_Char;

    procedure cmMouseLeave(var Msg: TMessage);     message cm_MouseLeave;
    procedure wmSetFocus(var Msg: TWMSetFocus);    message wm_SetFocus;
    procedure wmActivate(var Msg: TWMActivate);    message wm_Activate;
    procedure wmMouseActivate(var Msg: TWMMouseActivate); message wm_MouseActivate;
    procedure cmShowingChanged(var Msg: TMessage); message cm_ShowingChanged;

    procedure wmShowAnimated(var Msg: TMessage);   message wm_ShowAnimated;
    procedure wmHideSilent(var Msg: TMessage);     message wm_HideSilent;
    procedure wmKillAnimation(var Msg: TMessage);  message wm_KillAnimation;
    procedure wmKillTimer(var Msg: TMessage);      message wm_KillTimer;
    procedure wmInitState(var Msg: TMessage);      message wm_InitState;
    procedure wmSetFraming(var Msg: TMessage);     message wm_SetFraming;
    procedure wmSetDesigning(var Msg: TMessage);   message wm_SetDesigning;
    procedure wmSetEnabled(var Msg: TMessage);     message wm_SetEnabled;
    procedure wmSetFocusItem(var Msg: TMessage);   message wm_SetFocusItem;

    // utilities
    procedure SetSelectedIndex(Value: Integer);
    procedure OnPopupTimer(Sender: TObject);
    procedure SearchActiveMenuForShortcut(var Msg: TWMKey);
    procedure DestroySubMenuForm;
    procedure CreateSubMenuForm(Menu: TPopupMenu; Handle: HMenu; Items: TMenuItem);
    procedure RebuildToolTipWindow(Recreate: Boolean);
    procedure OnHotTrack(Sender: TObject);

    // bounds management
    procedure RebuildBounds;
    function GetItemRect(Index: Integer): TRect;
    function GetMenuItemHeight(const M: TMenuItem): Integer;
    function GetMenuItemHeightIndex(Index: Integer): Integer;
    function GetRealHeight: Integer; virtual;
    function GetRealWidth: Integer;  virtual;
    procedure GetOptions(Items: TMenuItem; Popup: TObject;
      var Options: T_AM2000_Options);
    procedure OnUnhideTimer(Sender: TObject);
    procedure AnimatedHide;
    function GetLastVisibleIndex: Integer;
    function GetMenuItemIndex1(Index: Integer; Items: TMenuItem): TMenuItem;
    procedure CheckOrigin(const CurAnim: T_AM2000_Animation; var NewLeft, NewTop, NewHeight: Integer);

    // painting
    procedure InitializePaintParams;
    procedure PaintBevelAndBorder;
    procedure PaintFrame3D;
    procedure PaintBackground;
    procedure PaintBackgroundPart(Y: Integer);
    procedure PaintDragPane;
    procedure PaintHiddenArrow;
    procedure PaintScrollBars;
    procedure PaintCachedDesktop(DC: HDC; BufferX, BufferY: Integer);

  protected
    procedure Paint; override;

    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
{$IFDEF Delphi4OrHigher}
    procedure DoClose(var Action: TCloseAction); override;
{$ENDIF}

  public
    CurMenuItem, MenuItems: TMenuItem;

    property SelectedIndex: Integer read FSelectedIndex write SetSelectedIndex;

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    function Animate: Boolean; virtual;
    procedure SilentShow;
    procedure SilentHide;
    procedure Repaint; override;

    procedure PopupSubMenuForm(SelectFirst: Boolean);
    function GetCurMenuItem: TMenuItem;
    function GetMenuItemIndex(Index: Integer): TMenuItem;

    // bounds management
    function GetIndexAt(X, Y: Integer): Integer;
    procedure OnScrollTimer(Sender: TObject);

    procedure Refresh;

  end;


  // forgot about this
  TFalseMenu = class(TComponent)
  private
{$IFDEF Delphi4OrHigher}
    FBiDiMode: TBiDiMode;
{$ENDIF}
    FItems: TMenuItem;
  end;

{ GetPopupMenuForm }

function GetPopupMenuForm(Item: TMenuItem): TCustomPopupMenu2000Form;
var
  F: TPopupMenu2000Form;
begin
  Result:= nil;
  if not AssignedActivePopupMenu2000Form
  then Exit;

  F:= TPopupMenu2000Form(TCustomPopupMenu2000(ActivePopupMenu).Form);
  while (F <> nil)
  and (F.MenuItems <> Item.Parent)
  do F:= F.SubMenuForm;

  Result:= F;
end;


{ TCustomPopupMenu2000 }

constructor TCustomPopupMenu2000.Create(AOwner: TComponent);
begin
  inherited;

  FFont:= TFont.Create;
  FParentFont:= False;
  FSystemFont:= True;
  FCtl3D:= True;
  FOptions:= T_AM2000_Options.Create;

  // we don't need TMenuItem so we just destoy it
  // and create TMenuItem2000 instead of.
  TFalseMenu(Self).FItems.Free;
  TFalseMenu(Self).FItems:= TMenuItem2000.Create(Self);
{$IFDEF Delphi4OrHigher}
  TFalseMenu(Self).FBiDiMode:= bdLeftToRight;
{$ENDIF}

  FEMI2000:= TEditableMenuItem2000.Create(Self);

  if (csDesigning in ComponentState)
  and (GetComponentItemsCaption <> '')
  then begin
    FComponentItem:= NewItem2000('::::' + GetComponentItemsCaption, '', False, False, nil, 0, '');
    Items.Add(FComponentItem);
  end;

  InstallWindowsHooks;
end;

destructor TCustomPopupMenu2000.Destroy;
begin
  RemoveWindowsHooks;

  if (FForm <> nil)
  and (Owner = nil)
  then FForm.Free;

  FFont.Free;
  FOptions.Free;
  FComponentItems.Free;

  PostMessage(GetMnuDsgnHandle, wm_Close, 0, 0);

  inherited;
end;

procedure TCustomPopupMenu2000.Popup(X, Y: Integer);
begin
  PopupAsync(X, Y);

  while (not (csDestroying in ComponentState))
  and FormOnScreen
  do begin
    Application.ProcessMessages;
    Sleep(20);
  end;
end;

procedure TCustomPopupMenu2000.PopupAsync(X, Y: Integer);
begin
  KillActivePopupMenu2000(False, False);
  ActivePopupMenu:= Self;

  // event
  if Assigned(OnPopup)
  then OnPopup(Self);

  // init items
  PrepareForPopup(False);

  // remove hidden items
  if not (PopupComponent is TCustomMenuBar2000)
  then RemoveShowHiddenFlag;

  with TPopupMenu2000Form(Form)
  do begin
    ParentMenuForm:= nil;
    AssignMenuItems(MenuItems, MenuHandle);

    // set animation options
    if (MenuItems <> nil)
    and (MenuItems is TMenuItem2000)
    and (TMenuItem2000(MenuItems).Options is T_AM2000_Options)
    then
      with T_AM2000_Options(TMenuItem2000(MenuItems).Options)
      do TPopupMenu2000Form(Form).Animation:= Animation

    else
      Animation:= FOptions.Animation;

    PopupMenu:= Self;
    Include(State, fsSelectedChanged);

    // show menu
    if (not IsWindowEnabled(Handle))
    then EnableWindow(Handle, True);

    PopupX:= X;
    PopupY:= Y;
    SetBounds(X, Y, 0, 0);

    // animation
    try
      if Animate then Exit;
    except
    end;

    SilentHide;
    if ActivePopupMenu = Self
    then ActivePopupMenu:= nil;
  end;
end;

procedure TCustomPopupMenu2000.AssignMenuItems(var DestItems: TMenuItem; var DestHandle: HMenu);
  // assigns menu items to TCustomPopupMenu2000Form
begin
  if MenuItems <> nil
  then DestItems:= Self.MenuItems
  else DestItems:= Self.Items2000;

  if MenuHandle <> 0
  then DestHandle:= Self.MenuHandle
  else DestHandle:= DestItems.Handle;
end;

function TCustomPopupMenu2000.FormOnScreen: Boolean;
begin
  Result:= (FForm <> nil)
    and (FForm.Height > 0)
    and (FForm.Width > 0)
    and (FForm.BorderStyle = bsNone);
end;

function TCustomPopupMenu2000.GetTopMostForm: TForm;
begin
  Result:= Form;
  while Assigned(TPopupMenu2000Form(Result).SubMenuForm)
  do Result:= TPopupMenu2000Form(Result).SubMenuForm;
end;

procedure TCustomPopupMenu2000.SetSelectedIndex(First: Boolean);
begin
  if FForm <> nil
  then
    with TPopupMenu2000Form(FForm) do begin
      if First
      then
        FSelectedIndex:= 0
      else
        FSelectedIndex:= itNothing;
        
      if FormOnScreen then Paint;
    end;
end;

function TCustomPopupMenu2000.IsFontStored: Boolean;
begin
  Result:= not (FParentFont or FSystemFont);
end;

function TCustomPopupMenu2000.IsShowHintStored: Boolean;
begin
  Result:= not FParentShowHint;
end;

function TCustomPopupMenu2000.GetMenuItem2000: TMenuItem2000;
begin
  Result:= TMenuItem2000(inherited Items);
end;

procedure TCustomPopupMenu2000.SetFont(Value: TFont);
begin
  FFont.Assign(Value);
end;

procedure TCustomPopupMenu2000.SetParentFont(Value: Boolean);
begin
  FParentFont:= Value;
  if Value then FSystemFont:= False;
end;

procedure TCustomPopupMenu2000.SetSystemFont(Value: Boolean);
begin
  FSystemFont:= Value;
  if Value then FParentFont:= False;
end;

procedure TCustomPopupMenu2000.SetShowHint(Value: Boolean);
begin
  if FShowHint <> Value then begin
    FShowHint:= Value;
    FParentShowHint:= False;
  end;
end;

procedure TCustomPopupMenu2000.SetParentShowHint(Value: Boolean);
begin
  if FParentShowHint <> Value then
    FParentShowHint:= Value;

  if FParentShowHint then
    if Owner is TControl
    then FShowHint:= TControl(Owner).ShowHint
    else FShowHint:= False;

end;

procedure TCustomPopupMenu2000.SetOptions(Value: T_AM2000_Options);
begin
  FOptions.Assign(Value);
end;

procedure TCustomPopupMenu2000.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;

  if (Operation = opRemove)
  then begin
{$IFNDEF Delphi4OrHigher}
    if (AComponent = FImages) then FImages:= nil;
{$ENDIF}
    if (AComponent = FStatusBar) then FStatusBar:= nil;
  end;
end;

procedure TCustomPopupMenu2000.PrepareForPopup(AddEmpty: Boolean);
var
  I: Integer;
  MI, MI2: TMenuItem2000;
begin
  if (FComponentItems <> nil)
  then begin
    // create basic item
    MI:= TMenuItem2000.Create(Owner);

    // create/update component items
    if FComponentItems.Count = 0
    then begin
      for I:= 0 to FComponentItems.Count -1
      do TMenuItem2000(FComponentItems[I]).Free;

      CreateComponentItems(MI, AddEmpty);
      
    end
    else begin
      for I:= 0 to FComponentItems.Count -1
      do begin
        MI2:= TMenuItem2000(FComponentItems[I]);
        MI2.Parent.Delete(MI2.MenuIndex);
        MI.Add(MI2);
      end;

      UpdateComponentItems(MI);
    end;

    // clear
    FComponentItems.Clear;

    // copy component items
    for I:= 0 to MI.Count -1
    do begin
      FComponentItems.Add(MI[0]);
      MI2:= MI[0];
      MI.Delete(0);
      FComponentItemsParent.Insert(FComponentItemsMenuIndex + I, MI2);
    end;
    
    // free basic menu item
    MI.Free;
  end;

  if (AddEmpty)
  and (Items.Count = 0)
  and (MenuItems = nil)
  and (MenuHandle = 0)
  then Items.Add(NewItem(SEmptyCaption, 0, False, False, nil, 0, ''));
end;

procedure TCustomPopupMenu2000.RemoveShowHiddenFlag;
begin
  if FForm <> nil
  then
    with TPopupMenu2000Form(Form) do
      if not (fsBecomingDraggable in State)
      then begin
        Exclude(State, fsShowHidden);
        Exclude(State, fsHiddenAsRegular);
      end;

  if not (mfHiddenAsRegular in Options.Flags)
  then Options.Flags:= Options.Flags - [mfHiddenIsVisible];
end;

function TCustomPopupMenu2000.IsShortCut(var Msg: TWMKey): Boolean;
begin
  Result:= IsShortCutEx(Msg, Items2000, csDesigning in ComponentState);
end;

procedure TCustomPopupMenu2000.Loaded;
var
  MI: TMenuItem2000;

  function FindItemByCaption(Items: TMenuItem2000; Caption: String): TMenuItem2000;
  var
    I: Integer;
  begin
    Result:= nil;

    for I:= 0 to Items.Count -1
    do begin
      if (Items[I].Caption = Caption)
      and (Items[I] <> FComponentItem)
      then begin
        Result:= Items[I];
        Exit;
      end;

      if Items[I].Count > 0
      then begin
        Result:= FindItemByCaption(Items[I], Caption);
        if Result <> nil
        then Exit;
      end;
    end;
  end;

begin
  inherited;

  if (GetComponentItemsCaption <> '')
  then begin
    MI:= FindItemByCaption(Items2000, '::::' + GetComponentItemsCaption);

    if (csDesigning in ComponentState)
    then begin
      // remove component item
      if MI <> nil
      then FComponentItem.Free;
    end
    else begin
      // remove component item
      if MI <> nil
      then begin
        FComponentItemsParent:= MI.Parent;
        FComponentItemsMenuIndex:= MI.MenuIndex;
      end
      else begin
        FComponentItemsParent:= Items2000;
        FComponentItemsMenuIndex:= 0;
      end;

      FComponentItems:= TList.Create;

      MI.Free;
    end;
  end;
end;

procedure TCustomPopupMenu2000.SetVersion(Value: T_AM2000_Version);
begin
end;

procedure TCustomPopupMenu2000.DoCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  if Assigned(FOnCloseQuery)
  then FOnCloseQuery(Sender, CanClose);
end;

function TCustomPopupMenu2000.GetComponentItemsCaption: String;
begin
  Result:= '';
end;

procedure TCustomPopupMenu2000.CreateComponentItems(Items: TMenuItem2000;
  AddEmpty: Boolean);
begin
  if (AddEmpty)
  and (Items.Count = 0)
  then Items.Add(NewItem(SEmptyCaption, 0, False, False, nil, 0, ''));
end;

procedure TCustomPopupMenu2000.UpdateComponentItems(Items: TMenuItem2000);
begin
end;

function TCustomPopupMenu2000.GetForm: TForm;
begin
  if FForm = nil
  then FForm:= TPopupMenu2000Form.Create(Owner);

  Result:= FForm;
end;

procedure TCustomPopupMenu2000.Refresh;
begin
  if FormOnScreen
  then TPopupMenu2000Form(Form).Refresh;
end;

function TCustomPopupMenu2000.GetVersion: T_AM2000_Version;
begin
  Result:= SVersion;
end;


{ TPopupMenu2000Form }

constructor TPopupMenu2000Form.Create(AOwner: TComponent);
begin
  inherited CreateNew(AOwner {$IFDEF CBuilder1}, 0{$ENDIF});

  // own initializaton
  ControlStyle:= ControlStyle - [csCaptureMouse];
  FSelectedIndex:= itNothing;
  CurHiddenCount:= 0;
  LastSelectedIndex:= itNothing;
  Animation:=  anVSlide;
  BorderStyle:= bsNone;
  FormStyle:= fsStayOnTop;
  Position:= poDesigned;

  Buffer:= TBitmap.Create;
  Buffer.Canvas.Font.Assign(Font);

  Back:= TBitmap.Create;

  ItemRects:= TList.Create;

  SetBounds(0, 0, 0, 0);
end;

destructor TPopupMenu2000Form.Destroy;
begin
  MenuItems:= nil;

  Back.Free;
  Buffer.Free;
  Options.Free;
  ItemRects.Free;
  DropShadow.Free;

  // free bits
  if bits <> nil then Freemem(bits);
  if bits0 <> nil then Freemem(bits0);
  if bits1 <> nil then Freemem(bits1);
  if dbits <> nil then Freemem(dbits);
  DeleteDC(DesktopDC);

  ToolTipWindow.Free;

  // this checking is necessary when submenu going to be killed
  // by MainForm (Owner property), not by ParenTCustomPopupMenu2000
  if (ParentMenuForm <> nil)
  and (ParentMenuForm.SubMenuForm = Self)
  then ParentMenuForm.SubMenuForm:= nil;

  if (SubMenuForm <> nil)
  then SubMenuForm.Release;

  if (ParentMenuForm = nil)
  and (PopupMenu <> nil)
  and (PopupMenu.FForm = Self)
  then PopupMenu.FForm:= nil;

  if FloatingMenusList <> nil
  then FloatingMenusList.Remove(Self);

  inherited Destroy;
end;


{ Properties }

function TPopupMenu2000Form.GetRealHeight: Integer;
begin
  Result:= GetItemRect(ItemRects.Count -1).Bottom +
    Options.Margins.Border +
    Options.Margins.Bevel;

  // options
  with Options do begin
    if (BackgroundDisplay = bdExpand)
    and (Result < Background.Height)
    then Result:= Background.Height;
  end;

  // hidden arrow
  if fsHiddenArrow in State
  then Inc(Result, iHiddenButtonHeight +6);

  // ctl3d
  if (BorderStyle = bsNone)
  and (fsCtl3d in State)
  then Inc(Result, 2);

end;

function TPopupMenu2000Form.GetRealWidth: Integer;
begin
  Result:= GetItemRect(0).Right +
    2*Options.Margins.Border +
    2*Options.Margins.Bevel;

  // options
  with Options do
    if (BackgroundDisplay = bdExpand)
    and (Result < Background.Width)
    then Result:= Background.Width;

  // ctl3d
  if (fsCtl3d in State)
  then Inc(Result, 2);

  // border <> bsNone
  if (BorderStyle <> bsNone)
  then Inc(Result, 2);

end;

function TPopupMenu2000Form.GetIndexAt(X, Y: Integer): Integer;
var
  I, H, B: Integer;
  P: TPoint;
begin
  P:= Point(X, Y);
  if (BorderStyle = bsNone)
  then
    if (fsCtl3D in State)
    then B:= Options.Margins.Bevel +2
    else B:= Options.Margins.Bevel
  else
    B:= 0;

  H:= Options.Margins.Border + B;

  Result:= itNothing;
  if (MenuHandle = 0)
  or (Y <= H)
  then Exit;

  // is title visible?
  with Options.Title
  do
    if (ParentMenuForm = nil)
    and Visible
    and (((Align = atLeft) and (X <= Width))
    or   ((Align = atRight) and (X >= Buffer.Width - Width)))
    then Exit;

  // is menu draggable?
  if (Options.Draggable)
  and (BorderStyle = bsNone)
  and (Y <= iDraggablePaneHeight +4)
  then begin
    Result:= itDragPane;
    Exit;
  end;


  // hidden arrow?
  if (fsHiddenArrow in State)
  and PtInRect(Rect(0, Buffer.Height - B - iHiddenButtonHeight, Buffer.Width,
    Buffer.Height - B), P)
  then begin
    Result:= itHiddenArrow;
    Exit;
  end;

  // check scrollbars
  if (fsScrollBars in State)
  then begin
    // top
    if (FirstVisibleIndex > 0)
    and PtInRect(Rect(0, B, Buffer.Width, B + iScrollButtonHeight), P)
    then begin
      Result:= itTopScroll;
      Exit;
    end;

    // bottom
    if (LastVisibleIndex < ItemRects.Count -1)
    and (PtInRect(Rect(0, Buffer.Height - B - iScrollButtonHeight, Buffer.Width,
      Buffer.Height - B), P))
    then begin
      Result:= itBottomScroll;
      Exit;
    end;
  end;

  // count all menu items
  if (FirstVisibleIndex > 0)
  then Inc(P.Y, GetItemRect(FirstVisibleIndex).Top - iScrollButtonHeight - B);

  // add framing
  if (fsFraming in State)
  then Inc(P.Y, ItemHeight div 2);

  for I:= FirstVisibleIndex to ItemRects.Count -1
  do
    if PtInRect(GetItemRect(I), P)
    then begin
      Result:= I;
      Exit;
    end;
end;

function TPopupMenu2000Form.GetLastVisibleIndex: Integer;
var
  H: Integer;
begin
  H:= MY;

  if (BorderStyle = bsNone)
  then
    if (fsCtl3D in State)
    then Dec(H, 2*Options.Margins.Border +4)
    else Dec(H, 2*Options.Margins.Border);

  if (fsScrollbars in State)
  and (FirstVisibleIndex > 0)
  then Inc(H, GetItemRect(FirstVisibleIndex).Top - iScrollButtonHeight);

  Dec(H, 10);

  Result:= FirstVisibleIndex +1;
  while (Result < ItemRects.Count)
  and (GetItemRect(Result).Bottom < H)
  do Inc(Result);
end;

procedure TPopupMenu2000Form.SetSelectedIndex(Value: Integer);
var
  Direction: Integer;
  M: TMenuItem;
begin
  if (Value <> SelectedIndex) then begin

    if Value > FSelectedIndex then Direction:= 1
                              else Direction:= -1;

    LastSelectedIndex:= FSelectedIndex;
    FSelectedIndex:= Value;
    Include(State, fsSelectedChanged);

    if MenuHandle = 0 then Exit;

    mii.fMask:= miim_Type + miim_State;
    repeat
      // in case if LastSelectedItem = itNothing
      if (FSelectedIndex = LastSelectedIndex) then Break;

      if FSelectedIndex < 0 then FSelectedIndex:= ItemRects.Count -1;
      if FSelectedIndex >= ItemRects.Count then begin
        FSelectedIndex:= itNothing;

        // show hidden menu items
        if fsHiddenArrow in State
        then begin
          StopTimer;
          Include(State, fsShowHidden);
          SilentHide;
          Animate;
          Exit;
        end;
      end;

      // no enabled items available
      if (FSelectedIndex = LastSelectedIndex)
      then Break;

      // try to find ordinal menu item
      M:= GetMenuItemIndex(FSelectedIndex);
      if (M <> nil)
      and (M.Caption <> SSeparator)
      and ((M.Enabled) or (not (mfNoHighDisabled in Options.Flags)))
      then Break;

      // Win32 menu item
      if (M = nil)
      and (FSelectedIndex >= 0)
      then begin
        mii.dwTypeData:= Z;
        mii.cch:= SizeOf(Z) -1;
        if GetMenuItemInfo(MenuHandle, FSelectedIndex, True, mii)
        and (mii.fType and mft_Separator = 0)
        and (mii.fState and (mfs_Grayed + mfs_Disabled) = 0)
        then Break;
      end;

      Inc(FSelectedIndex, Direction);

    until False;

    Include(State, fsDrawDisabled);
    Paint;

    CheckShowHint(GetCurMenuItem, False, Self);

    // set context
    if (CurMenuItem <> nil)
    then HelpContext:= CurMenuItem.HelpContext;
  end;
end;

procedure TPopupMenu2000Form.GetOptions(Items: TMenuItem; Popup: TObject; var Options: T_AM2000_Options);
  // returns options for the current submenu
var
  Item: TMenuItem2000;
begin
  Options.Free;
  Options:= T_AM2000_Options.Create;

  // assign parent options
  if (Popup <> nil)
  and (Popup is TCustomPopupMenu2000)
  then Options.Assign(TCustomPopupMenu2000(Popup).Options);

  // title cannot be inherited
  if ParentMenuForm <> nil
  then Options.Title.Visible:= False;

  // assign current options
  if (Items is TMenuItem2000)
  then begin
    Item:= TMenuItem2000(Items);
    while (Item.Parent <> nil)
    and (Item.Parent is TMenuItem2000)
    and (Item.Options.IsDefault)
    do Item:= TMenuItem2000(Item.Parent);

    if not Item.Options.IsDefault
    then Options.Assign(Item.Options);
  end;
end;



{ Events }

procedure TPopupMenu2000Form.WMActivate(var Msg: TWMActivate);
var
  F: Forms.TForm;
begin
  if (Msg.Active = wa_Active)
  and (Owner is Forms.TForm)
  then begin
    F:= Forms.TForm(Owner);
    while (F.FormStyle = fsMdiChild)
    and (F.Owner is Forms.TForm)
    do F:= Forms.TForm(F.Owner);

    if (F is Forms.TForm)
    and (F.Visible)
    and not (csDestroying in F.ComponentState)
    then SendMessage(F.Handle, wm_NCActivate, 1, 0);

  end
  else
    inherited;

end;

// many thanks to Jordan Russell again!...
procedure TPopupMenu2000Form.WMMouseActivate(var Msg: TWMMouseActivate);
begin
  Msg.Result:= ma_NoActivate;
  if (Owner is Forms.TForm) then
    SetActiveWindow(Forms.TForm(Owner).Handle);
end;

procedure TPopupMenu2000Form.CMMouseLeave(var Msg: TMessage);
var
  C: TControl;
  P: TPoint;
begin
  inherited;

  MouseState:= [];
  if (fsAnimated in State)
  or (fsDisabled in State)
  then Exit;

  // no menu or no submenu
  if not (HasSubMenu(GetCurMenuItem)
  or HasSubMenuWin32(MenuHandle, FSelectedIndex))
  then begin
    LastSelectedIndex:= FSelectedIndex;
    FSelectedIndex:= itNothing;
    Include(State, fsSelectedChanged);
    Paint;
  end;

  // moving items
  if fsFraming in State
  then begin
    InsertBeforeItem:= miNothingToInsert;
    InsertBeforeParent:= nil;
    Repaint;

    GetCursorPos(P);
    C:= FindDragTarget(P, True);
    if (C = nil)
    or not IsOkControl(C)
    then Screen.Cursor:= crDragReject;
  end;

  // clear status bar
  SetStatusBarText('');

  // hottrack
  if (PopupMenu <> nil)
  and (PopupMenu.HotTrack)
  then begin
    GetCursorPos(P);
    C:= FindDragTarget(P, True);
    if (C = nil)
    or not IsOkControl(C)
    then SetTimer(nHotTrackTimeout, OnHotTrack);
  end;
end;

procedure TPopupMenu2000Form.WMSetFocus(var Msg: TWMSetFocus);
begin
  if (fsAnimated in State)
  and (Owner is Forms.TForm)
  and (Forms.TForm(Owner).Visible)
  then Forms.TForm(Owner).SetFocus
  else inherited;
end;

procedure TPopupMenu2000Form.wmSetFraming(var Msg: TMessage);
begin
  if Msg.wParam <> 0
  then begin
    Include(State, fsFraming);
    SetDragCursor;
  end
  else begin
    Exclude(State, fsFraming);
    RestoreCursor;
  end;

  if SubMenuForm <> nil
  then SubMenuForm.wmSetFraming(Msg);

  Msg.Result:= 1;
end;

procedure TPopupMenu2000Form.CMShowingChanged(var Msg: TMessage);
begin
  // Skip Application.UpdateVisible
  // that shows annoying 'application icon' on taskbar
  if Showing
  then begin
    ShowWindow(Handle, sw_ShowNA);
    SetWindowPos(Handle, hwnd_Top, Left, Top, Width, Height, FormFlags);
  end
  else
    ShowWindow(Handle, sw_Hide);
end;

procedure TPopupMenu2000Form.MouseDown;
var
  I: Integer;
  F: TPopupMenu2000Form;
begin
  if ssLeft  in Shift then Include(MouseState, msLeftButton);
  if ssRight in Shift then Include(MouseState, msRightButton);
  Include(State, fsMouseChanged);

  // hide tooltip
  if (ToolTipWindow <> nil)
  then ToolTipWindow.Deactivate;

  // click on dragpane
  I:= GetIndexAt(X, Y);
  if (Options.Draggable)
  and (I = itDragPane)
  and (Button = mbLeft)
  then begin
    // hide previous floating menu
    for I:= 0 to FloatingMenusList.Count -1
    do
      with TPopupMenu2000Form(FloatingMenusList[I])
      do
        if MenuItems = Self.MenuItems
        then begin
          FloatingMenusList.Delete(I);
          Close;
          Break;
        end;

    F:= TPopupMenu2000Form.Create(Owner);

    // Draggable pane clicked!
    if Options.Caption <> ''
    then F.Caption:= Options.Caption
    else
    if MenuItems.Caption <> ''
    then F.Caption:= StripAmpersands(MenuItems.Caption)
    else F.Caption:= 'Animated Menus';

    // set window properties
    F.BorderStyle:= bsToolWindow;
    F.MenuItems:= MenuItems;
    F.MenuHandle:= MenuHandle;
    F.Animation:= anNone;
    F.PopupMenu:= PopupMenu;
    F.Options:= T_AM2000_Options.Create;
    F.Options.Assign(Options);

    Include(F.State, fsBecomingDraggable);
    Include(F.State, fsShowHidden);
    Include(F.State, fsHiddenAsRegular);

    // rebuild bounds
    F.RebuildBounds;

    F.MX:= F.GetRealWidth;
    F.MY:= F.GetRealHeight;
    if SystemParametersInfo(spi_GetNonClientMetrics, 0, @NonClientMetrics, 0)
    then Inc(F.MY, Integer(NonClientMetrics.iCaptionHeight) +3);

    F.SetBounds(Left, Top, F.MX, F.MY);
    F.Buffer.Width:= F.MX;
    F.Buffer.Height:= F.MY;
    F.Buffer.FreeImage;

    F.LastVisibleIndex:= F.GetLastVisibleIndex;

    // paint menu
    F.Paint;

    KillActivePopupMenu2000(True, False);
    F.Show;
    F.Refresh;

    SendMessage(F.Handle, WM_SysCommand, $F012, 0);  // system message

    // tooltip window
    F.RebuildToolTipWindow(True);

    Exclude(F.State, fsBecomingDraggable);
    FloatingMenusList.Add(F);
  end


  // bottom scroll
  else if (SelectedIndex = itBottomScroll)
  then begin
    if LastVisibleIndex < ItemRects.Count -1
    then begin
      Inc(FirstVisibleIndex);
      LastVisibleIndex:= GetLastVisibleIndex;
      Repaint;

      if not IsTimerSet(nScrollingTimeout, OnScrollTimer)
      then begin
        StopTimer;
        SetTimer(nScrollingTimeout, OnScrollTimer);
      end;

      Exit;
    end;
  end

  // top scroll
  else if (SelectedIndex = itTopScroll)
  then begin
    if FirstVisibleIndex > 0
    then begin
      Dec(FirstVisibleIndex);
      LastVisibleIndex:= GetLastVisibleIndex;
      Repaint;

      if not IsTimerSet(nScrollingTimeout, OnScrollTimer)
      then begin
        StopTimer;
        SetTimer(nScrollingTimeout, OnScrollTimer);
      end;

      Exit;
    end;
  end

  // ordinary paint
  else begin
    if fsDisabled in State
    then begin
      Exclude(State, fsDisabled);
      DestroySubMenuForm;
    end;

    Paint;
    ActiveMenuItem:= GetMenuItemIndex(LastSelectedIndex);
  end;
end;

procedure TPopupMenu2000Form.MouseUp;
var
  wID: Integer;
  MI: TMenuItem;
  CanClose: Boolean;
  P: TPoint;
  hWnd: THandle;
  F: TPopupMenu2000Form;
//  F: TPopupMenu2000Form;

  procedure CheckHiddenSeparator(Index, Delta: Integer; Items: TMenuItem);
    // look for first not-a-hidden menu item
    // if the separator exists between current menu item and first not-a-hidden
    // then promote the separator too
  var
    M, SeparatorItem: TMenuItem;
    mic: Integer;
  begin
    // exit if hidden items are hidden
    if not (fsShowHidden in State) then Exit;

    SeparatorItem:= nil;
    mic:= ItemRects.Count;
    repeat
      Index:= Index + Delta;

      // exit if no not-a-hidden item found
      if (Index < 0)
      or (Index >= mic)
      then Exit;

      M:= GetMenuItemIndex1(Index, Items);

      // exit if no not-a-hidden item found
      if (M = nil)
      and (Index < mic)
      then Exit;

      // separator found
      if (M <> nil)
      and (SeparatorItem = nil)
      and (M.Caption <> '')
      and (M.Caption[1] = SSeparator)
      then SeparatorItem:= M;

      // hidden menu item
      if (M is TMenuItem2000)
      and TMenuItem2000(M).Hidden
      then Continue;

      // not-a-hidden item found!
      if (SeparatorItem <> nil)
      and (SeparatorItem is TMenuItem2000)
      and ((M = nil) or (M.Caption[1] <> SSeparator))
      then begin
        TMenuItem2000(SeparatorItem).Hidden:= False;
        SaveMenuItemToRegistry(SeparatorItem, False);
      end;

      Exit;
    until False;
  end;

  procedure PromoteMenuItem(MI: TMenuItem);
    // clear hidden flag at menu item and all parents
  begin
    if (MI is TMenuItem2000)
    and (TMenuItem2000(MI).Hidden)
    and (EnablePromotion)
    then begin
      TMenuItem2000(MI).Hidden:= False;
      SaveMenuItemToRegistry(MI, False);

      // and show separators
      CheckHiddenSeparator(LastSelectedIndex, -1, MI.Parent);
      CheckHiddenSeparator(LastSelectedIndex, 1, MI.Parent);
    end;

    // check parents
    if MI <> nil
    then
      PromoteMenuItem(MI.Parent);
  end;

begin
  if not (ssLeft  in Shift) then Exclude(MouseState, msLeftButton);
  if not (ssRight in Shift) then Exclude(MouseState, msRightButton);
  Include(State, fsMouseChanged);

  StopTimer;
  
  // exit when animated
  if (fsAnimated in State)
  or (fsKillAnimate in State)
  then Exit;

  if fsIgnoreNextMouseUp in State
  then begin
    Exclude(State, fsIgnoreNextMouseUp);
    Exit;
  end;

  // if hidden arrow is pressed
  if SelectedIndex = itHiddenArrow
  then begin
    Include(State, fsShowHidden);
    SilentHide;
    Animate;
    Exit;
  end;

  // moving menu items
  if fsFraming in State
  then begin
    // end moving
    Exclude(State, fsFraming);
    EndCustomization(True);

    // resizing
    RebuildBounds;
    MX:= GetRealWidth;
    MY:= GetRealHeight;
    SetBounds(Left, Top, MX, MY);
    Buffer.Width:= MX;
    Buffer.height:= MY;
    LastVisibleIndex:= GetLastVisibleIndex;

    // repaint
    Repaint;
    Exit;
  end;

  // nothing is selected
  if (SelectedIndex = itNothing)
  or (SelectedIndex = itTopScroll)
  or (SelectedIndex = itBottomScroll)
  then begin
    Paint;
    Exit;
  end;

  // popup context menu
  MI:= GetCurMenuItem;
  if (Button = mbRight)
  then begin
    if (MI <> nil)
    and (MI is TMenuItem2000)
    then begin
      if Assigned(Self.PopupMenu.OnContextMenu)
      then Self.PopupMenu.OnContextMenu(MI);

      if TMenuItem2000(MI).PopupMenu <> nil
      then
        with TMenuItem2000(MI)
        do begin
          Include(State, fsDisabled);
          DestroySubMenuForm;

          // on popup event
          if Assigned(PopupMenu.OnPopup)
          then PopupMenu.OnPopup(MI);

          // create context menu as submenu
          CreateSubMenuForm(PopupMenu, PopupMenu.Items.Handle, PopupMenu.Items);
          PopupMenu.PopupComponent:= MI;
          SubMenuForm.Animation:= Self.Options.Animation;

          // show context menu
          GetCursorPos(P);
          SubMenuForm.SetBounds(P.X, P.Y, 0, 0);
          SubMenuForm.Animate;
      end;
    end;
    
    Exit;
  end;

  // cancel editing
  if (fsDisabled in State)
  and (FocusItem <> nil)
  and (SelectedIndex <> GetIndexAt(X, Y))
  then begin
    TMenuItem2000(FocusItem).AsEdit.CancelEdit(True);
    Exclude(State, fsDisabled);
    FocusItem:= nil;
    Paint;
    Exit;
  end;

  // clear status bar
  SetStatusBarText('');

  if SubMenuForm <> nil
  then begin
    Exclude(State, fsDisabled);
    PopupSubMenuForm(False);
    Exit;
  end;


  CanClose:= True;
  if (MI <> nil)
  and (MenuItems.IndexOf(MI) <> -1)
  then begin

    // ordinary menu item
    if HasSubmenu(MI)
    then begin
      PopupSubMenuForm(False);
      Exit;
    end;
    if (not MI.Enabled) or (MI.Caption = SSeparator) then Exit;

    // is editbox
    if (MI is TMenuItem2000)
    and (TMenuItem2000(MI).Control = ctlEditbox)
    then begin
      TMenuItem2000(MI).AsEdit.BeginEdit(X, Y);
      Exit;
    end;

    // clear hidden flag at MI and it's parents
    PromoteMenuItem(MI);

    // check OnCloseQuery
    PopupMenu.DoCloseQuery(MI, CanClose);

    if CanClose
    then begin
      // set clicked state
{      F:= Self;
      while F <> nil
      do begin
        Include(F.State, fsClicked);
        F:= F.ParentMenuForm;
      end;
}
      // close the menu
      // check for popup menu
      F:= ParentMenuForm;
      while (F <> nil)
      and (F.PopupMenu = PopupMenu)
      do F:= F.ParentMenuForm;

      if F = nil
      then KillActivePopupMenu2000(True, True)
      else begin
        Exclude(F.State, fsDisabled);
        F.SubMenuForm.Release;
      end;

      FullShowCaret;
      if MenuSoundResource
      then PlaySound(PChar(MenuCommandSound), HInstance, snd_Resource + snd_Async + snd_NoDefault + snd_NoStop)
      else sndPlaySound(PChar(MenuCommandSound), snd_Async + snd_NoDefault + snd_NoStop);
    end;

    if Assigned(PopupMenu.OnMenuCommand)
    then PopupMenu.OnMenuCommand(MI);

    // set item index for button array
    if (MI is TMenuItem2000)
    and (TMenuItem2000(MI).Control = ctlButtonArray)
    then
      with TMenuItem2000(MI).AsButtonArray do
        ItemIndex:= LastItemIndex;

    // Click
    MI.Click;
  end
  else begin

    mii.fMask:= miim_Type + miim_ID + miim_Submenu;
    mii.dwTypeData:= Z;
    mii.cch:= SizeOf(Z) -1;
    if (not GetMenuItemInfo(MenuHandle, SelectedIndex + CurHiddenCount, True, mii))
    or (mii.fType and mft_Separator <> 0)
    or (mii.fState and mfs_Disabled <> 0)
    then Exit;

    if mii.hSubmenu <> 0
    then begin
      PopupSubMenuForm(False);
      Exit;
    end;

    wID:= mii.wID;
    StopTimer;
    PopupMenu.DoCloseQuery(MI, CanClose);

    if CanClose
    then begin
      // set clicked state
{      F:= Self;
      while F <> nil
      do begin
        Include(F.State, fsClicked);
        F:= F.ParentMenuForm;
      end;
}
      // close the menu
      KillActivePopupMenu2000(True, True);
      
      if Owner is TForm
      then SendMessage(TForm(Owner).Handle, wm_NCActivate, 1, 0);
      FullShowCaret;

      if MenuSoundResource
      then PlaySound(PChar(MenuCommandSound), HInstance, snd_Resource + snd_Async + snd_NoDefault + snd_NoStop)
      else sndPlaySound(PChar(MenuCommandSound), snd_Async + snd_NoDefault + snd_NoStop);
    end;

    if Assigned(PopupMenu.OnMenuCommand)
    then PopupMenu.OnMenuCommand(nil);

    // send message
    if (PopupMenu.Ole2ObjectWindow <> 0)
    then
      if PopupMenu.MenuPopulated
      then begin
        MI:= FindItem2000(TForm(Owner), wID, fkCommand);
        if (MI <> nil)
        and MI.Enabled
        then
          MI.Click;
      end
      else
        PostMessage(PopupMenu.Ole2ObjectWindow, wm_Command, wID, 0)

    else begin

      if Owner is TForm
      then hWnd:= TForm(Owner).Handle
      else
      if PopupMenu.PopupComponent.Owner is TForm
      then hWnd:= TForm(PopupMenu.PopupComponent.Owner).Handle
      else hWnd:= 0;

      PostMessage(hWnd, wm_Command, wID, 0)
    end;
  end;

  // show must go on
  if (not CanClose)
  then begin
    Paint;

    if (ToolTipWindow <> nil)
    then ToolTipWindow.Activate;

    SetTimer(nPopupSubmenuTimeout, OnPopupTimer);
  end;
end;

procedure TPopupMenu2000Form.MouseMove(Shift: TShiftState; X, Y: Integer);
const
  LastX: Integer = 0;
  LastY: Integer = 0;
var
  R: TRect;
  I, L: Integer;
begin
  Include(MouseState, msMouseOver);

  if (fsAnimated in State)
  or (fsKillAnimate in State)
  or (fsDisabled in State)
  then Exit;

  // relay mouse move
  if (ToolTipWindow <> nil)
  then ToolTipWindow.RelayMouseMove(SmallPoint(X, Y));

  // change mouse cursor
  if (GetCurMenuItem is TMenuItem2000)
  and (TMenuItem2000(CurMenuItem).Control = ctlEditbox)
  then begin
    if (BorderStyle = bsNone)
    then
      if (fsCtl3D in State)
      then L:= Options.Margins.Border +2
      else L:= Options.Margins.Border
    else
      L:= 0;

    Inc(L, AmpTextWidth(Buffer.Canvas, CurMenuItem.Caption) +5);

    R:= GetItemRect(FSelectedIndex);
    if R.Left < L then R.Left:= L;
    InflateRect(R, 0, -2);

    if PtInRect(R, Point(X, Y))
    then
      Cursor:= crIBeam
    else
      if Cursor <> crDefault
      then Cursor:= crDefault;

  end
  else
    if Cursor = crIBeam
    then Cursor:= crDefault;


  // ignore some little movements
  if ((X = LastX) and (Y = LastY))
  then Exit;

  LastX:= X;
  LastY:= Y;

  if fsIgnoreMouseMove in State
  then begin
    Exclude(State, fsIgnoreMouseMove);
    Exit;
  end;

  if IsTimerSet(nPopupSubmenuTimeout, OnPopupTimer)
  or IsTimerSet(nShortPopupSubmenuTimeout, OnPopupTimer)
  then StopTimer;
  
  // set parentmenuform highlight
  if (ParentMenuForm <> nil)
  and (not (fsDisabled in ParentMenuForm.State))
  and (ParentMenuForm.FSelectedIndex <> ParentMenuIndex)
  then
    with ParentMenuForm
    do begin
      LastSelectedIndex:= FSelectedIndex;
      FSelectedIndex:= Self.ParentMenuIndex;
      Include(State, fsSelectedChanged);
      Exclude(State, fsDrawDisabled);
      Paint;
    end;

  // set highlight
  I:= GetIndexAt(X, Y);
  if I <> SelectedIndex
  then begin
    if not (fsDesigning in State)
    then begin
      LastSelectedIndex:= FSelectedIndex;
      FSelectedIndex:= I;
      Include(State, fsSelectedChanged);
      Exclude(State, fsDrawDisabled);
    end;

    // start moving
    if IsCustomization(Options.Customizable)
    then begin
      if (SelectedIndex <> -1)
      or (LastSelectedIndex <> 0)
      then InsertBeforeItem:= GetMenuItemIndex(SelectedIndex);

      if (InsertBeforeItem = nil)
      then
        if (SelectedIndex = itHiddenArrow)
        or (SelectedIndex = itBottomScroll)
        or PtInRect(GetItemRect(ItemRects.Count -1), Point(X, Y))
        then
          InsertBeforeItem:= miInsertAtTheEnd
        else begin
          Screen.Cursor:= crDragReject;
          InsertBeforeItem:= miNothingToInsert;
          Paint;
          Exit;
        end;

      InsertBeforeParent:= MenuItems;
      Include(State, fsFraming);
      if PopupMenu.PopupComponent is TCustomMenuBar2000
      then PostMessage(TCustomMenuBar2000(PopupMenu.PopupComponent).Handle, wm_SetFraming, 1, 0);

      SetDragCursor;
    end;

    // repaint
    Paint;

    // show menu item hint
    SetStatusBarText('');
    CheckShowHint(GetCurMenuItem, True, Self);

    // is it hidden arrow?
    if (SelectedIndex = itHiddenArrow)
    then begin
      StopTimer;
      DestroySubmenuForm;
      SetTimer(nShowHiddenTimeout, OnUnhideTimer);
      Exit;
    end;
  end
  else
    if (GetCurMenuItem <> nil)
    and (CurMenuItem is TMenuItem2000)
    and (TMenuItem2000(CurMenuItem).Control = ctlButtonArray)
    then
      with TMenuItem2000(CurMenuItem), AsButtonArray do
        if GetIndexAt(X, Y) <> LastItemIndex
        then begin
          Self.Paint;
          CheckShowHint(GetCurMenuItem, True, Self);
        end;

  if (SubMenuForm <> nil)
  or (HasSubMenu(CurMenuItem))
  then
    if (SubMenuForm <> nil)
    and (HasSubMenu(CurMenuItem))
    then SetTimer(nShortPopupSubmenuTimeout, OnPopupTimer)
    else SetTimer(nPopupSubmenuTimeout, OnPopupTimer);
end;

procedure TPopupMenu2000Form.SearchActiveMenuForShortcut(var Msg: TWMKey);
  // just a symbol key - seeking for menu item
var
  I: Integer;
  M: TMenuItem;
  S: String;
begin
  for I:= 0 to ItemRects.Count -1
  do begin
    M:= GetMenuItemIndex(I);
    if (M = nil) then begin
      mii.fMask:= miim_ID + miim_Type + miim_State;
      mii.dwTypeData:= @Z;
      mii.cch:= SizeOf(Z) -1;
      if (not GetMenuItemInfo(MenuHandle, I, True, mii))
      or (mii.fType and mft_Separator <> 0)
      or (mii.fState and (mfs_Disabled + mfs_Grayed) <> 0)
      then Continue;

      S:= StrPas(Z);
    end
    else
      if (not (M.Visible and M.Enabled))
      or ((M is TMenuItem2000)
      and (TMenuItem2000(M).Hidden)
      and (not (fsShowHidden in State)))
      then Continue
      else S:= M.Caption;

    if IsAccel2(Msg.CharCode, S)
    then begin
      SelectedIndex:= I;
      Exclude(State, fsSelectedChanged);

      if ((M <> nil) and (M.Enabled))
      or ((M = nil) and (mii.fState and (mfs_Grayed + mfs_Disabled) = 0))
      then MouseUp(mbLeft, [], 0, 0)
      else MessageBeep(0);

      Msg.Result:= 1;
      Exit;
    end;
  end { for  };

  MessageBeep(0);
end;

procedure TPopupMenu2000Form.wmKeyDown(var Msg: TWMKeyDown);
  // keyboard events
var
  M: TMsg;
  Shift: TShiftState;
begin
  // handle focus item
  if (FocusItem <> nil)
  and (fsDisabled in State)
  and (FocusItem is TMenuItem2000)
  then begin
    // get shift state
    Shift:= [];
    if GetKeyState(vk_Shift) < 0    then Include(Shift, ssShift);
    if GetKeyState(vk_Control) < 0  then Include(Shift, ssCtrl);
    if Msg.KeyData and AltMask <> 0 then Include(Shift, ssAlt);

    // handle char
    case Msg.CharCode of
      vk_Escape, vk_Return:
        begin
          TMenuItem2000(FocusItem).AsEdit.CancelEdit(Msg.CharCode = vk_Return);
          Exclude(State, fsDisabled);
          FocusItem:= nil;
          Paint;
          Msg.Result:= 1;
        end;

      else
        // send char
        TMenuItem2000(FocusItem).ControlOptions.KeyDown(Msg.CharCode, Shift);
    end;

    // set result
    if Msg.CharCode = 0 then Msg.Result:= 1;
  end

  // other menu item
  else

  case Msg.CharCode of
    // click item
    vk_Return:
      begin
        MouseUp(mbLeft, [], 0, 0);
        Msg.Result:= 1;
      end;

    vk_Up:
      begin
        SelectedIndex:= SelectedIndex -1;

        if (SelectedIndex > LastVisibleIndex)
        or (SelectedIndex < FirstVisibleIndex)
        then begin
          if (SelectedIndex < LastVisibleIndex)
          then begin
            Dec(FirstVisibleIndex);
            LastVisibleIndex:= GetLastVisibleIndex;
          end
          else
            while LastVisibleIndex < SelectedIndex do begin
              Inc(FirstVisibleIndex);
              LastVisibleIndex:= GetLastVisibleIndex;
            end;

          Repaint;
        end;

        StopTimer;
        Msg.Result:= 1;
      end;

    vk_Tab:
      begin
        if GetKeyState(vk_Shift) < 0
        then SelectedIndex:= SelectedIndex -1
        else SelectedIndex:= SelectedIndex +1;
        
        StopTimer;
        Msg.Result:= 1;
      end;

    vk_Down:
      begin
        if GetKeyState(vk_Control) < 0
        then begin
          StopTimer;
          Include(State, fsShowHidden);
          SilentHide;
          Animate;
        end
        else begin
          SelectedIndex:= SelectedIndex +1;

          if (SelectedIndex > LastVisibleIndex)
          or (SelectedIndex < FirstVisibleIndex)
          then begin
            if (SelectedIndex > LastVisibleIndex)
            then Inc(FirstVisibleIndex)
            else FirstVisibleIndex:= SelectedIndex;

            LastVisibleIndex:= GetLastVisibleIndex;
            Repaint;
          end;

          StopTimer;
        end;
        Msg.Result:= 1;
      end;

    vk_Left:
      if ParentMenuForm <> nil then begin;
        if not (fsAnimated in State)
        then SilentHide;

        Release;
        Msg.Result:= 1;
      end;

    vk_Right:
      if (GetCurMenuItem <> nil)
      and (CurMenuItem.Enabled)
      and (HasSubmenu(CurMenuItem)
      or HasSubMenuWin32(MenuHandle, FSelectedIndex))
      then begin
        PopupSubMenuForm(True);
        Msg.Result:= 1;
      end;

    vk_Escape:
      begin
        if fsFraming in State
        then begin
          // cancel moving
          Exclude(State, fsFraming);
          Include(State, fsSelectedChanged);
          Include(State, fsMouseChanged);
          Include(State, fsIgnoreNextMouseUp);

          EndCustomization(False);

          Repaint;
        end
        else begin
          if not (fsAnimated in State)
          then SilentHide;

          if (ParentMenuForm <> nil)
          then begin
            Exclude(ParentMenuForm.State, fsDisabled);
            Release;
          end
          else
            KillActivePopupMenu2000(False, False);
        end;

        Msg.Result:= 1;
      end;

    vk_F1:
      if (Application.HelpFile <> '')
      and (GetCurMenuItem <> nil)
      and (CurMenuItem.HelpContext <> 0)
      then begin
        KillActivePopupMenu2000(True, False);
        if (PopupMenu <> nil)
        and (PopupMenu.Owner is TForm)
        and (biHelp in TForm(PopupMenu.Owner).BorderIcons)
        then Application.HelpCommand(HELP_CONTEXTPOPUP, CurMenuItem.HelpContext)
        else Application.HelpCommand(HELP_CONTEXT, CurMenuItem.HelpContext);

        // We have to remove the next message that is in the queue.
        PeekMessage(M, 0, 0, 0, pm_Remove);
        Msg.Result:= 1;
      end;

    else
      if (Msg.CharCode <> vk_Shift) and (Msg.CharCode <> vk_Control)
      then begin
        SearchActiveMenuForShortcut(Msg);
        Msg.Result:= 1;
      end;
  end;

  // hide tooltip
  if (ToolTipWindow <> nil)
  then ToolTipWindow.Deactivate;
end;

procedure TPopupMenu2000Form.wmChar(var Msg: TWMChar);
var
  C: Char;
begin
  // handle focus item
  if (FocusItem <> nil)
  and (fsDisabled in State)
  then begin
    // send char
    C:= Char(Msg.CharCode);

    if (FocusItem is TMenuItem2000)
    then TMenuItem2000(FocusItem).ControlOptions.KeyPress(C);

    // set result
    if C = #0 then Msg.Result:= 1;
  end;
end;

procedure TPopupMenu2000Form.wmSysKeyDown(var Msg: TWMKeyDown);
begin
  // hide active menu
  if (Msg.CharCode = vk_Menu) or (Msg.CharCode = vk_F10) then begin
    KillActivePopupMenu2000(True, False);

    if (PopupMenu <> nil)
    and (PopupMenu.PopupComponent is TCustomMenuBar2000)
    then TCustomMenuBar2000(PopupMenu.PopupComponent).SetDisableAltKeyUp(True);

    FullShowCaret;
    Msg.Result:= 1;
    Exit;
  end;

  // handle as menu key
  wmKeyDown(Msg);
end;


procedure TPopupMenu2000Form.OnPopupTimer(Sender: TObject);
begin
  StopTimer;
end;

procedure TPopupMenu2000Form.OnUnhideTimer(Sender: TObject);
begin
  StopTimer;

  // check is mouse cursor still over the hidden arrow
  if SelectedIndex = itHiddenArrow then begin
    Include(State, fsShowHidden);
    Animate;
  end;
end;

procedure TPopupMenu2000Form.OnScrollTimer(Sender: TObject);
var
  P: TPoint;
begin
  if not IsTimerSet(nShortScrollingTimeout, OnScrollTimer)
  then begin
    StopTimer;
    SetTimer(nShortScrollingTimeout, OnScrollTimer);
  end;
  
  GetCursorPos(P);
  MouseDown(mbLeft, [ssLeft], P.X, P.Y);
end;

procedure TPopupMenu2000Form.DestroySubMenuForm;
begin
  if SubMenuForm = nil then Exit;

  SubMenuForm.SilentHide;
  SubMenuForm.Release;
  SubMenuForm:= nil;
end;

procedure TPopupMenu2000Form.CreateSubMenuForm(Menu: TPopupMenu; Handle: HMenu; Items: TMenuItem);
begin
end;

procedure TPopupMenu2000Form.PopupSubMenuForm(SelectFirst: Boolean);
begin
  // destroy previous submenu form
  DestroySubmenuForm;
end;

function TPopupMenu2000Form.GetCurMenuItem: Menus.TMenuItem;
begin
  Result:= GetMenuItemIndex(SelectedIndex);
end;

function TPopupMenu2000Form.GetMenuItemIndex1(Index: Integer; Items: TMenuItem): TMenuItem;
var
  I, C: Integer;
  M: TMenuItem;
begin
  Result:= nil;
  if (Items = nil)
  or (Index < 0)
  or (Index >= MenuItems.Count)
  then Exit;

  CurHiddenCount:= 0;
  C:= -1;
  for I:= 0 to Items.Count -1 do begin
    M:= Items[I];

    // menu item is not visible
    if not M.Visible then Continue;

    // menu item is hidden
    if  (not (fsShowHidden in State))
    and (M is TMenuItem2000)
    and (TMenuItem2000(M).Hidden)
    then begin
      Inc(CurHiddenCount);
      Continue;
    end;

   // else
    Inc(C);
    if C = Index then begin
      Result:= M;
      Exit;
    end;
  end;
end;

function TPopupMenu2000Form.GetMenuItemIndex(Index: Integer): TMenuItem;
begin
  Result:= GetMenuItemIndex1(Index, MenuItems);
  CurMenuItem:= Result;
end;



{ Main Routines }

procedure TPopupMenu2000Form.CheckOrigin(const CurAnim: T_AM2000_Animation; var NewLeft, NewTop, NewHeight: Integer);
const
  SM_CXVIRTUALSCREEN = 78;
  SM_CYVIRTUALSCREEN = 79;

var
  X, ScreenWidth, ScreenHeight, OldTop: Integer;

  function GetAppropriateHeight(AHeight: Integer): Integer;
  var
    I, H, B, L: Integer;
  begin
    H:= AHeight - iScrollButtonHeight +1;

    if (BorderStyle = bsNone)
    then
      if (fsCtl3D in State)
      then B:= 2*Options.Margins.Border +4
      else B:= 2*Options.Margins.Border
    else
      B:= 0;

    Dec(H, B);

    I:= ItemRects.Count -1;
    L:= GetItemRect(I).Bottom;
    while (I > 1)
    and (L - GetItemRect(I -1).Top < H)
    do Dec(I);

    Result:= L - GetItemRect(I).Top + iScrollButtonHeight + B +1;
  end;

begin
  NewLeft:= Left;
  NewTop:= Top;

  // calculate desktop's size without TaskBar
  x:= GetSystemMetrics(sm_CxScreen);
  ScreenWidth:= GetSystemMetrics(sm_CxVirtualScreen);
  if x > ScreenWidth then ScreenWidth:= x;

  x:= GetSystemMetrics(sm_CyScreen);
  ScreenHeight:= GetSystemMetrics(sm_CyVirtualScreen);
  if x > ScreenHeight then ScreenHeight:= x;

  ///////////////////
  // horizontal
  ///////////////////

  Exclude(State, fsFromRightToLeft);

  // change the direction for long submenus
  if (ParentMenuForm <> nil)
  and (fsFromRightToLeft in ParentMenuForm.State)
  and (ParentMenuForm.Left - MX +6 > 0)
  then begin
    NewLeft:= ParentMenuForm.Left +6;
    Include(State, fsFromRightToLeft);
  end;

  // out of right side
  if (Left + MX) > ScreenWidth
  then begin
    Include(State, fsFromRightToLeft);

    if ParentMenuForm = nil
    then begin
      // topmenu
      if (PopupMenu.PopupComponent is TCustomMenuBar2000)
      or (PopupMenu.PopupComponent is TButton)
      or (PopupMenu.PopupComponent is TSpeedButton)
      then begin
        NewLeft:= ScreenWidth - MX;
        Exclude(State, fsFromRightToLeft);
      end;
      
    end
    else begin
      // submenu
      NewLeft:= ParentMenuForm.Left +6;
    end;
  end;

  // final
  if NewLeft < 0 then NewLeft:= 0;


  ///////////////////
  // vertical
  ///////////////////

  Exclude(State, fsFromBottomToTop);

  // out of bottom side
  if (Top + MY) > ScreenHeight
  then begin
    if ParentMenuForm = nil
    then begin
      OldTop:= NewTop;

      // menu shouldn't hide a menu button
      if (PopupMenu.PopupComponent is TCustomMenuBar2000)
      then
        with TCustomMenuBar2000(PopupMenu.PopupComponent) do
          if mbType = mbVertical
          then
            NewTop:= ScreenHeight
          else
            NewTop:= NewTop - (aiRect.Bottom - aiRect.Top +1);

      // menu shouldn't hide other buttons
      if (PopupMenu.PopupComponent is TButton)
      then
        with TButton(PopupMenu.PopupComponent) do
          NewTop:= NewTop - Height -1;

      if (PopupMenu.PopupComponent is TSpeedButton)
      then
        with TSpeedButton(PopupMenu.PopupComponent) do
          NewTop:= NewTop - Height -1;

      // can we fit the menu into the upper part of the screen?
      if OldTop > MY
      then // yes
        Include(State, fsFromBottomToTop)

      else begin // no
        NewTop:= OldTop;
        Include(State, fsScrollBars);
        
        MY:= GetAppropriateHeight(ScreenHeight - NewTop);
        if not (CurAnim in [anVSlide, anUnfold])
        then NewHeight:= MY;
      end;

    end
    else begin
      if MY < ScreenHeight
      then
        NewTop:= ScreenHeight - MY

      else begin
        MY:= GetAppropriateHeight(ScreenHeight);
        NewTop:= ScreenHeight - MY;
        Include(State, fsScrollBars);

        if not (CurAnim in [anVSlide, anUnfold])
        then NewHeight:= MY;
      end;
    end;
  end;

  // final
  if NewTop < 0 then NewTop:= 0;

end;

function TPopupMenu2000Form.Animate: Boolean;
var
  CurAnim: T_AM2000_Animation;
  X, NewLeft, NewTop, NewWidth, NewHeight: Integer;
begin
  Perform(wm_InitState, 0, 0);

  DX:= 1;
  DY:= 1;
  Result:= True;
  CurAnim:= anNone;
  CurHiddenCount:= 0;
  FocusItem:= nil;
  Include(State, fsAnimated);
  Exclude(State, fsKillAnimate);
  if nFirstStage < 1 then nFirstStage:= 1;


  // calculate normal height and width.
  // mx and mx -- width and height of the full opened window
  RebuildBounds;
  MX:= GetRealWidth;
  MY:= GetRealHeight;

  // top, left -- real coordinates (click)
  // width, height = 0
  // mx, my -- real width and height
  // newleft, newtop, newheight, newwidth -- what to show next

  // code for calculation animation steps
  if (PopupMenu = nil)
  or (ParentMenuForm <> nil)
  or (not PopupMenu.IgnoreAnimation)
  then begin
    CurAnim:= Animation;
    if CurAnim = anRandom
    then begin
      if Random(2) = 0
      then CurAnim:= anHSlide
      else CurAnim:= anVSlide;
    end;

    if CurAnim = anSmart
    then begin
      Animation:= anHSlide;
      CurAnim:= anVSlide;
    end;

    if CurAnim = anPopup
    then begin
      Animation:= anHSlide;
      CurAnim:= anUnfold;
    end;

    if CurAnim in [anHSlide, anUnfold] then DX:= MX div nSteps;
    if DX < 3 then DX:= 3;
    if CurAnim in [anVSlide, anUnfold] then DY:= MY div nSteps;
    if DY < 3 then DY:= 3;

    if DX < DY then DX:= DY;

    if CurAnim in [anHSlide, anUnfold]
    then NewWidth:= MX div nFirstStage
    else NewWidth:= MX;

    if CurAnim in [anVSlide, anUnfold]
    then NewHeight:= MY div nFirstStage - DY
    else NewHeight:= MY;

  end
  else begin
    NewWidth:= MX;
    NewHeight:= MY;
    if Animation in [anSmart, anPopup]
    then Animation:= anHSlide;
  end;


  // code to arrange menu form
  FirstVisibleIndex:= 0;
  LastVisibleIndex:= 0;

  // check origin
  CheckOrigin(CurAnim, NewLeft, NewTop, NewHeight);

  ///////////////////
  // final settings
  ///////////////////

  if (fsFromBottomToTop in State)
  then NewTop:= NewTop - NewHeight;

  if (fsFromRightToLeft in State)
  then NewLeft:= NewLeft - NewWidth;

  SetBounds(NewLeft, NewTop, NewWidth, NewHeight);

  // correct VisibleItemsCount
  if fsScrollbars in State
  then begin
    LastVisibleIndex:= GetLastVisibleIndex;
  end
  else
    LastVisibleIndex:= ItemRects.Count -1;

  // initialize buffer
  Buffer.Width:= MX;
  Buffer.Height:= MY;
  Buffer.FreeImage;


  //////////////////////
  // act!
  //////////////////////

  // check if mouse already moved
  ProcessMouseMoveMessages;
  if (fsKillAnimate in State)
  then begin
    Exclude(State, fsAnimated);
    Result:= False;
    Exit;
  end;

  // start
  if dbits <> nil then Freemem(bits);
  dbits:= nil;

  Include(State, fsPaintMenu);
  Paint;

  // init fade in animation
  if (CurAnim = anFadeIn)
  then begin
    // restore last values
    if bits0 <> nil
    then begin
      SetDIBits(Buffer.Canvas.Handle, Buffer.Handle, 0, MY, bits0, bi, Dib_Rgb_Colors);
      BitBlt(DesktopDC, Left, Top, Buffer.Width, Buffer.Height, Buffer.Canvas.Handle, 0, 0, SrcCopy);
    end;

    BitsSize:= (MX +1) * MY * SizeOf(TRgbTriple);

    // nado
    if bits1 <> nil then Freemem(bits1);
    GetMem(bits1, BitsSize);

    bi.bmiHeader.biSize:= SizeOf(bi.bmiHeader);
    bi.bmiHeader.biWidth:= MX;
    bi.bmiHeader.biHeight:= MY;
    bi.bmiHeader.biPlanes:= 1;
    bi.bmiHeader.biBitCount:= 24;
    bi.bmiHeader.biCompression:= bi_RGB;
    bi.bmiHeader.biSizeImage:= 0;
    GetDIBits(Buffer.Canvas.Handle, Buffer.Handle, 0, MY, bits1, bi, Dib_Rgb_Colors);

    // create desktop DC
    if DesktopDC = 0
    then DesktopDC:= CreateDC('DISPLAY', nil, nil, nil);

    // est
    BitBlt(Buffer.Canvas.Handle, 0, 0, MX, MY, DesktopDC, Left, Top, SrcCopy);

    if bits0 <> nil then Freemem(bits0);
    GetMem(bits0, BitsSize);
    GetDIBits(Buffer.Canvas.Handle, Buffer.Handle, 0, MY, bits0, bi, Dib_Rgb_Colors);

    // anFadeIn
    // temp
    if bits <> nil then Freemem(bits);
    GetMem(bits, BitsSize);

    // iterate
    GetMem(dbits, BitsSize);
    for X:= 0 to BitsSize -1 do
      dbits^[X]:= ShortInt((ShortInt(bits1^[X]) - ShortInt(bits0^[X]))) div (nSteps);

    CurStep:= 1;
  end;

  if CurAnim <> anFadeIn
  then SilentShow;

  if MenuSoundResource
  then PlaySound(PChar(MenuPopupSound), HInstance, snd_Resource + snd_Async + snd_NoDefault + snd_NoStop)
  else sndPlaySound(PChar(MenuPopupSound), snd_Async + snd_NoDefault + snd_NoStop);

  SendMessage(Handle, wm_ShowAnimated, GetCurrentTime, nTimeout);
end;

procedure TPopupMenu2000Form.wmKillAnimation(var Msg: TMessage);
  // cancels active animation
begin
  if fsBecomingDraggable in State then Exit;
  Include(State, fsKillAnimate);
  Exclude(State, fsAnimated);
  Exclude(State, fsIgnoreNextMenuUp);

  // process submenu
  if (SubMenuForm <> nil)
  then SubMenuForm.Perform(wm_KillAnimation, 0, 0);
end;

procedure TPopupMenu2000Form.wmShowAnimated(var Msg: TMessage);
var
  CT, X, NewLeft, NewTop, NewWidth, NewHeight: Integer;
begin
  if (fsKillAnimate in State) then Exit;

  // delay
  CT:= Msg.LParam - (GetCurrentTime - Msg.WParam);
  if (CT > 0) and (CT < 1000) then Sleep(CT);

  NewLeft:= Left;
  NewTop:= Top;
  NewWidth:= Width;
  NewHeight:= Height;
  Msg.LParam:= 0;

  if Animation = anFadeIn
  then begin
    if CurStep <= nSteps
    then begin
      TimeStart:= GetCurrentTime;

      for X:= 0 to BitsSize -1 do
        bits^[X]:= bits0^[X] + dbits^[X] * CurStep;

      SetDIBits(Buffer.Canvas.Handle, Buffer.Handle, 0, MY, bits, bi, Dib_Rgb_Colors);
      BitBlt(DesktopDC, Left, Top, Buffer.Width, Buffer.Height, Buffer.Canvas.Handle, 0, 0, Canvas.CopyMode);

      ProcessMouseMoveMessages;

      Inc(CurStep);
      CT:= GetCurrentTime;
      PostMessage(Handle, wm_ShowAnimated, CT, nTimeout - (CT - TimeStart));

      Exit;
    end;

    Include(State, fsPaintMenu);
    SetBounds(NewLeft, NewTop, NewWidth, NewHeight);
    SetDIBits(Buffer.Canvas.Handle, Buffer.Handle, 0, MY, bits1, bi, Dib_Rgb_Colors);
    SilentShow;
    BitBlt(Canvas.Handle, 0, 0, Buffer.Width, Buffer.Height, Buffer.Canvas.Handle, 0, 0, Canvas.CopyMode);

    ProcessPaintMessages;

  end
  else begin
    // Animation!
    while (Width < MX) or (Height < MY)
    do begin
      // calculate the time wasted on drawing
      TimeStart:= GetCurrentTime;

      if NewWidth < MX then
        if (NewWidth + DX) >= MX
        then Break
        else begin
          NewWidth:= NewWidth + DX;
          if fsFromRightToLeft in State
          then NewLeft:= NewLeft - DX;
        end;

      if NewHeight < MY then
        if (NewHeight + DY) >= MY
        then Break
        else begin
          NewHeight:= NewHeight + DY;
          if fsFromBottomToTop in State
          then NewTop:= NewTop - DY;
        end;

      SetBounds(NewLeft, NewTop, NewWidth, NewHeight);
      Paint;

      if DropShadow <> nil
      then DropShadow.Refresh;

      ProcessMouseMoveMessages;
      ProcessPaintMessages;

      // adjust delay
      CT:= GetCurrentTime;
      PostMessage(Handle, wm_ShowAnimated, CT, nTimeout - (CT - TimeStart));

      Exit;
    end;

    // final setting bounds
    if fsFromBottomToTop in State
    then NewTop:= Top - (MY - Height) +1;

    if fsFromRightToLeft in State
    then NewLeft:= Left - (MX - Width);
  end;

  // finish
  SetBounds(NewLeft, NewTop, MX, MY);
  Paint;

  if DropShadow <> nil
  then DropShadow.Refresh;

  // finalization
  Exclude(State, fsAnimated);
  Exclude(State, fsKillAnimate);

  if Assigned(PopupMenu)
  and (SelectedIndex <> -1)
  then SetTimer(nPopupSubmenuTimeout, OnPopupTimer);
end;

procedure TPopupMenu2000Form.SilentShow;
var
  TopMostForm: TPopupMenu2000Form;
  LastLeft, LastTop: Integer;
begin
  // special check for tray-icon menu
  FullHideCaret;
  RebuildToolTipWindow(False);

  Exclude(State, fsDisabled);

  Color:= Options.Colors.Menu;

  if (mfDropShadow in Options.Flags)
  then begin
    if (DropShadow = nil)
    then DropShadow:= T_AM2000_DropShadowForm.Create(Self);

    LastLeft:= Self.Left;
    if (fsFromRightToLeft in State)
    then Dec(LastLeft, MX - Width);

    LastTop:= Self.Top;
    if (fsFromBottomToTop in State)
    then Dec(LastTop, MY - Height);

    DropShadow.GetDesktop(LastLeft, LastTop, MX, MY,
      (PopupMenu <> nil) and (PopupMenu.KeepDesktop));
    DropShadow.DirectionX:= not (fsFromRightToLeft in State);
    DropShadow.DirectionY:= not (fsFromBottomToTop in State);
    ShowWindow(DropShadow.Handle, sw_ShowNA);
  end
  else
  if DropShadow <> nil
  then DropShadow.Free;

  ShowWindow(Handle, sw_ShowNA);
  if Assigned(Application.MainForm)
  and not Application.MainForm.Visible
  then SetForegroundWindow(Handle);

  if Owner is TForm
  then SetForegroundWindow(TForm(Owner).Handle);

  if DropShadow <> nil
  then
    with DropShadow
    do begin
      TopMostForm:= Self;
      while TopMostForm.ParentMenuForm <> nil
      do TopMostForm:= TopMostForm.ParentMenuForm;

      SetWindowPos(Handle, TopMostForm.Handle, Self.Left, Self.Top, 0, 0, FormFlags);
      Refresh;
    end;

  SetWindowPos(Handle, hwnd_Top, Left, Top, Width, Height, FormFlags);
end;

procedure TPopupMenu2000Form.SilentHide;
  // hides the current window
begin
  StopTimer;

  if SubMenuForm <> nil
  then begin
    SubMenuForm.SilentHide;
    SubMenuForm.Release;
    SubMenuForm:= nil;
  end;

  if DropShadow <> nil
  then
    with DropShadow
    do begin
      ShowWindow(Handle, sw_Hide);
      SetBounds(Left, Top, 0, 0);
      ClearDesktop;
    end;

  if fsBecomingDraggable in State
  then begin
    // false alarm
    SetTimer(nPopupSubmenuTimeout, OnPopupTimer);
    Exit;
  end;

  if (Width <> 0)
  or (Height <> 0)
  then begin
    ShowWindow(Handle, sw_Hide);
    SetBounds(Left, Top, 0, 0);
  end;

  FSelectedIndex:= itNothing;

  // restore desktop
  if Animation = anFadeIn
  then begin
    SetDIBits(Buffer.Canvas.Handle, Buffer.Handle, 0, MY, bits0, bi, Dib_Rgb_Colors);
    BitBlt(DesktopDC, Left, Top, Buffer.Width, Buffer.Height, Buffer.Canvas.Handle, 0, 0, SrcCopy);
  end;

  ProcessPaintMessages;

  // remove arrays
  if dbits <> nil then FreeMem(dbits); dbits:= nil;
  if bits <> nil then FreeMem(bits);  bits:= nil;
  if bits0 <> nil then FreeMem(bits0);  bits0:= nil;
  if bits1 <> nil then FreeMem(bits1); bits1:= nil;

  // play close sound
  if MenuSoundResource
  then PlaySound(PChar(MenuCloseSound), HInstance, snd_Resource + snd_Async + snd_NoDefault + snd_NoStop)
  else sndPlaySound(PChar(MenuCloseSound), snd_Async + snd_NoDefault + snd_NoStop);

  // removing too tips
  if ToolTipWindow <> nil
  then begin
    ToolTipWindow.Free;
    ToolTipWindow:= nil;
  end;

  // emptying background
  ClearBitmap(Back);

  // fire OnMenuClose
end;

procedure TPopupMenu2000Form.AnimatedHide;
  // hides the current window
var
  I, X: Integer;
begin
  StopTimer;

  if (SelectedIndex >= 0)
  and (SelectedIndex < ItemRects.Count)
  then begin
    // result
    SetDIBits(Buffer.Canvas.Handle, Buffer.Handle, 0, MY, bits0, bi, Dib_Rgb_Colors);

    if ParentMenuForm <> nil
    then ParentMenuForm.PaintCachedDesktop(Buffer.Canvas.Handle, Left, Top);
    GetDIBits(Buffer.Canvas.Handle, Buffer.Handle, 0, MY, bits0, bi, Dib_Rgb_Colors);

    // est
    Paint;
    GetDIBits(Buffer.Canvas.Handle, Buffer.Handle, 0, MY, bits1, bi, Dib_Rgb_Colors);
    Move(bits1^, bits^, BitsSize);

    // temp
    for X:= 0 to BitsSize -1 do
      dbits^[X]:= Integer(Integer(bits1^[X]) - Integer(bits0^[X])) div (nSteps);

    // animation
    for I:= nSteps -1 downto 0 do begin
      for X:= 0 to BitsSize -1 do
        if dbits^[X] <> 0
        then bits^[X]:= bits0^[X] + dbits^[X] * I;

      SetDIBits(Buffer.Canvas.Handle, Buffer.Handle, 0, MY, bits, bi, Dib_Rgb_Colors);
      BitBlt(DesktopDC, Left, Top, Buffer.Width, Buffer.Height, Buffer.Canvas.Handle, 0, 0, SrcCopy);
      Sleep(2 * nTimeout);
    end;
  end;

  SilentHide;
end;

procedure TPopupMenu2000Form.Repaint;
begin
  Include(State, fsPaintMenu);
  Paint;
end;

procedure TPopupMenu2000Form.WMKillTimer(var Msg: TMessage);
begin
  StopTimer;

  // process submenu
  if (SubMenuForm <> nil)
  then SubMenuForm.Perform(wm_KillTimer, 0, 0);
end;

procedure TPopupMenu2000Form.wmHideSilent(var Msg: TMessage);
begin
  if (Msg.LParam <> 0)
  and (Animation = anFadeIn)
  and (SubMenuForm = nil)
  then
    AnimatedHide

  else begin
    // process drop shadows
    if (DropShadow <> nil)
    then ShowWindow(DropShadow.Handle, sw_Hide);

    ShowWindow(Handle, sw_Hide);

    // process submenus
    if (SubMenuForm <> nil)
    then SubMenuForm.Perform(wm_HideSilent, 0, Msg.LParam);

    SilentHide;
  end;
end;


{ Get Routines }

function TPopupMenu2000Form.GetMenuItemHeight(const M: TMenuItem): Integer;
var
  Lines, ImagesHeight: Integer;
begin
  Result:= 0;
  if M = nil then Exit;

  if (M.Caption <> '')
  and (M.Caption[1] = SSeparator)
  then begin
    if M.Caption = SSeparator
    then Result:= Options.SeparatorHeight
    else Result:= Options.SeparatorHeight +2;
  end

  else begin
    Lines:= GetNumLines(M.Caption);
    Inc(Result, (ItemHeight -3) * Lines +3 + 2*Options.Margins.Frame);

    if (M is TMenuItem2000)
    then
      with TMenuItem2000(M) do
      begin
        if (Control <> ctlNone)
        then
          Result:= TMenuItem2000(M).GetHeight(Buffer.Canvas, Self.Options)
        else
          if Hidden
          and not (fsHiddenAsRegular in State)
          then Inc(Result, 2);

        if (not Bitmap.Empty)
        and (Result < Bitmap.Height +4)
        then Result:= Bitmap.Height +4;

        if (DefaultIndex >= 0)
        and (Result < 20)
        then Result:= 20;

        if (M.Parent <> nil)
{$IFNDEF Delphi5OrHigher}
        and (M.Parent is TMenuItem2000)
        and (TMenuItem2000(M.Parent).SubMenuImages <> nil)
{$ELSE}
        and (M.Parent.SubMenuImages <> nil)
{$ENDIF}
        then
          ImagesHeight:= TMenuItem2000(M.Parent).SubMenuImages.Height

        else
        if Self.PopupMenu.Images <> nil
        then
          ImagesHeight:= Self.PopupMenu.Images.Height
        else
          ImagesHeight:= 0;

        if (ImageIndex >= 0)
        and (Result < ImagesHeight +4)
        then Result:= ImagesHeight +4;

      end;
  end;
end;

function TPopupMenu2000Form.GetMenuItemHeightIndex(Index: Integer): Integer;
begin
  with GetItemRect(Index) do Result:= Bottom - Top;
end;


{ Painting }

procedure TPopupMenu2000Form.InitializePaintParams;
begin
  with PaintParams, mir
  do begin
    State:= [];

    Canvas:= Buffer.Canvas;
    Handle:= Self.MenuHandle;
    Options:= Self.Options;
    MouseState:= Self.MouseState;
    FullRepaint:= fsPaintMenu in Self.State;

    // set ParentFont
    if (Options.Colors.MenuText = clMenuText)
    then Options.Colors.MenuText:= Font.Color;

    // show accels
    SetState(State, isShowAccelerators, (PopupMenu <> nil) and (PopupMenu.ShowAccelerators));
    SetState(State, isCtl3d, (fsCtl3d in Self.State));

    if (MenuItems <> nil)
    and (MenuItems.Parent <> nil)
  {$IFDEF Delphi5OrHigher}
    and (MenuItems.Parent.SubMenuImages <> nil)
  {$ELSE}
    and (MenuItems.Parent is TMenuItem2000)
    and (TMenuItem2000(MenuItems.Parent).SubMenuImages <> nil)
  {$ENDIF}
    then
      Images:= TImageList(TMenuItem2000(MenuItems.Parent).SubMenuImages)
    else
      if (PopupMenu <> nil)
      then
        Images:= TImageList(TCustomPopupMenu2000(PopupMenu).Images);

    // calculate bounds
    if (BorderStyle <> bsNone)
    then begin
      Border:= 0;
      LineRight:= ClientWidth;
    end
    else begin
      if (fsCtl3D in Self.State)
      then Ctl3DWidth:= 2
      else Ctl3DWidth:= 0;

      Border:= Ctl3DWidth +
        Options.Margins.Border +
        Options.Margins.Bevel;
      Width:= Buffer.Width;
      LineRight:= Buffer.Width - Border;
    end;

    LineLeft:= Border;

    ItemWidth:= Self.ItemWidth;
    ShortcutWidth:= Self.ShortcutWidth;

    if Options.Title.Visible
    then
      with mir
      do
        case Options.Title.Align of
          atLeft:
            begin
              LeftTitleWidth:= Options.Title.Width;
              TitleWidth:= LeftTitleWidth;
              Inc(LineLeft, LeftTitleWidth);
              if BorderStyle <> bsNone
              then Inc(LineLeft, 1);
            end;
          atRight:
            begin
              LeftTitleWidth:= 0;
              TitleWidth:= Options.Title.Width;
              Dec(LineRight, TitleWidth);
              if BorderStyle <> bsNone
              then Dec(LineRight, 1);
            end;
        end

    else begin
      LeftTitleWidth:= 0;
      TitleWidth:= 0;
    end;

    if Options.Alignment <> taRightToLeft
    then begin
      ItemLeft:= LineLeft + Options.Margins.Left;
      ShortcutLeft:= LineRight - Options.Margins.Right - ShortcutWidth;
      BitmapLeft:= LineLeft;
      TriangleWidth:= Options.Margins.Right -4;
      TriangleLeft:= LineRight - TriangleWidth -2;
    end
    else begin
      ShortcutLeft:= LineLeft + Options.Margins.Left;
      ItemLeft:= LineRight - Options.Margins.Right - itemWidth;
      BitmapLeft:= LineRight - ItemHeight -1;
      TriangleWidth:= Options.Margins.Left -4;
      TriangleLeft:= Border;
    end;

    if Options.Alignment <> taRightToLeft
    then BitmapWidth:= LineLeft + Options.Margins.Left - BitmapLeft -3
    else BitmapWidth:= LineRight - BitmapLeft -2;
  end;
end;

procedure TPopupMenu2000Form.PaintFrame3D;
var
  R: TRect;
begin
  with Buffer, Canvas
  do begin
    R:= ClipRect;

    if (BorderStyle = bsNone)
    and (fsCtl3D in State)
    then DrawEdge(Handle, R, bdr_RaisedInner + bdr_RaisedOuter, bf_Rect);
  end;
end;

procedure TPopupMenu2000Form.PaintBevelAndBorder;
var
  R: TRect;
  X, XLeft, XRight: Integer;
  M: TMenuItem;
begin
  with Buffer, Canvas, PaintParams.mir, Options.Margins
  do begin
    if Back.Empty
    then begin
      if Brush.Style <> bsSolid
      then Brush.Style:= bsSolid;

      // border
      if Brush.Color <> Options.Colors.Border
      then Brush.Color:= Options.Colors.Border;
      if Pen.Color <> Options.Colors.Border
      then Pen.Color:= Options.Colors.Border;

      X:= Bevel + Ctl3dWidth;
      XLeft:= X + LeftTitleWidth;
      XRight:= X + TitleWidth - LeftTitleWidth;

      FillRect(Rect(XLeft, X, XLeft + Border, Buffer.Height - X)); // left
      FillRect(Rect(Buffer.Width - XRight - Border, X, Buffer.Width - XRight, Buffer.Height - X)); // right

      M:= GetMenuItemIndex(FirstVisibleIndex);
      if (M is TMenuItem2000)
      and (TMenuItem2000(M).Hidden)
      then
        if Options.Colors.Sunken <> clNone
        then Brush.Color:= Options.Colors.Sunken
        else Brush.Bitmap:= Pattern;

      FillRect(Rect(XLeft, X, Buffer.Width - XRight, X + Border)); // top

      M:= GetMenuItemIndex(LastVisibleIndex);
      if (M is TMenuItem2000)
      and (TMenuItem2000(M).Hidden)
      then
        if Options.Colors.Sunken <> clNone
        then Brush.Color:= Options.Colors.Sunken
        else Brush.Bitmap:= Pattern

      else
        Brush.Color:= Options.Colors.Border;
        
      FillRect(Rect(XLeft, Buffer.Height - X - Border, Buffer.Width - XRight, Buffer.Height - X)); // bottom

      // bevel
      if Bevel > 0
      then begin
        R:= Rect(0, 0, Buffer.Width, Buffer.Height);
        if (BorderStyle = bsNone)
        and (fsCtl3D in State)
        then InflateRect(R, -2, -2);

        Frame3D(Buffer.Canvas,
          R,
          Options.Colors.Bevel,
          Options.Colors.BevelShadow,
          Bevel);
      end;
    end
    else begin
      // paint top part of the background bitmap
      BitBlt(Handle, Ctl3dWidth + LeftTitleWidth,
        Ctl3dWidth,
        Width - 2*Ctl3dWidth - TitleWidth,
        Bevel + Border,
        Back.Canvas.Handle, 0, 0, SrcCopy);
    end;

    // paint title
    if Options.Title.Visible
    then begin
      Options.Title.Paint(Buffer.Canvas, (BorderStyle = bsNone), ClientOrigin.Y - Self.Top);
    end;
  end;
end;

procedure TPopupMenu2000Form.PaintBackground;
var
  R: TRect;
  DC: HDC;
  hb: HBrush;
begin
  // fill background when empty
  if not Back.Empty
  then Exit;

  DC:= GetWindowDC(Handle);
  Back.Handle:= CreateCompatibleBitmap(DC, MX, MY);
  ReleaseDC(Handle, DC);

  if ((Options.Background.Width < MX -2)
  or (Options.Background.Height < MY -2))
  and (Options.BackgroundDisplay <> bdTile)
  then begin
    DC:= Back.Canvas.Handle;
    hb:= CreateSolidBrush(ColorToRGB(Options.Colors.Menu));
    FillRect(DC, Rect(0, 0, MX, MY), hb);
    DeleteObject(hb);
  end;

  case Options.BackgroundDisplay of
    bdDefault, bdExpand:
      begin
      BitBlt(Back.Canvas.Handle, 0, 0, MX, MY,
        Options.Background.Canvas.Handle, 0, 0, SrcCopy);
      end;

    bdCenter:
      with PaintParams.mir
      do begin
        R.Left:= LineLeft;
        R.Right:= R.Left + Options.Background.Width;
        R.Top:= Border;
        R.Bottom:= R.Top + Options.Background.Height;

        OffsetRect(R, ((LineRight - LineLeft) - Options.Background.Width) div 2 - Border,
          (Buffer.Height - Options.Background.Height) div 2 - Border);

        BitBlt(Back.Canvas.Handle, R.Left, R.Top, R.Right - R.Left, R.Bottom - R.Top,
          Options.Background.Canvas.Handle, 0, 0, SrcCopy);
      end;

    bdTile:
      with Back.Canvas do begin
        Brush.Bitmap:= Options.Background;
        FillRect(ClipRect);
      end;

    bdStretch:
      with PaintParams.mir, Options.Background do
        StretchBlt(Back.Canvas.Handle, 0, 0, Buffer.Width, Buffer.Height,
          Canvas.Handle, 0, 0, Width, Height, SrcCopy);
  end;
end;

procedure TPopupMenu2000Form.PaintBackgroundPart(Y: Integer);
begin
  with Buffer, Canvas, PaintParams.mir
  do begin
    // draw the bottom part of background picture
    if (Y < Buffer.Height)
    then begin
      if not Options.Background.Empty
      then begin
        with PaintParams.mir, Options.Margins
        do
          BitBlt(Handle, Ctl3dWidth + LeftTitleWidth, Y,
            Width - 2*Ctl3dWidth - TitleWidth, Buffer.Height - Y - Ctl3dWidth,
            Back.Canvas.Handle, 0, Y, Buffer.Canvas.CopyMode)
      end
      else begin
        if Brush.Color <> Options.Colors.Menu
        then Brush.Color:= Options.Colors.Menu;
        if Brush.Style <> bsSolid
        then Brush.Style:= bsSolid;

        FillRect(Rect(PaintParams.mir.LineLeft, Y,
          PaintParams.mir.LineRight, Buffer.Height - Border));
      end;
    end;
  end;
end;

procedure TPopupMenu2000Form.PaintDragPane;
var
  Y1, C1, C2, R1, G1, B1, I1, I2: Integer;
  DR, DG, DB, DH, F1, F2: Real;
  M: TMenuItem;
  R: TRect;
begin
  with Buffer, Canvas do begin
    Brush.Style:= bsSolid;

    if SelectedIndex = itDragPane
    then begin C1:= ColorToRGB(clActiveCaption); C2:= ColorToRGB(clGradientActiveCaption); end
    else begin C1:= ColorToRGB(clInactiveCaption); C2:= ColorToRGB(clGradientInactiveCaption); end;

    if IsGradientCaptions
    then begin
      // gradient!
      DH:= (PaintParams.mir.LineRight - PaintParams.mir.LineLeft)/256;
      R1:= GetRValue(C1);
      G1:= GetGValue(C1);
      B1:= GetBValue(C1);
      DR:= (GetRValue(C2) - R1)/255;
      DG:= (GetGValue(C2) - G1)/255;
      DB:= (GetBValue(C2) - B1)/255;

      for Y1:= 0 to 255
      do begin
        Brush.Color:= Rgb(R1 + Round(DR*Y1), G1 + Round(DG*Y1), B1 + Round(DB*Y1));
        F1:= Y1*DH;
        F2:= (Y1 +1)*DH;
        I1:= Round(F1);
        I2:= Round(F2);
        FillRect(Rect(PaintParams.mir.LineLeft + I1, PaintParams.mir.Border,
          PaintParams.mir.LineLeft + I2, PaintParams.mir.Border + iDraggablePaneHeight));
      end;
    end

    else begin
      Brush.Color:= C1;
      FillRect(Rect(PaintParams.mir.LineLeft, PaintParams.mir.Border, PaintParams.mir.LineRight, PaintParams.mir.Border +7));
    end;

    // bottom line
    R:= Rect(PaintParams.mir.LineLeft, PaintParams.mir.Border +7, PaintParams.mir.LineRight, PaintParams.mir.Border +10);

    M:= GetMenuItemIndex(FirstVisibleIndex);
    if (M is TMenuItem2000)
    and (TMenuItem2000(M).Hidden)
    then Dec(R.Bottom);

    Brush.Color:= Options.Colors.Menu;
    FillRect(R);

  end;
end;

procedure TPopupMenu2000Form.PaintHiddenArrow;
var
  R: TRect;
  X, Y: Integer;
begin
  with Buffer, Canvas, PaintParams.mir
  do begin
    R:= Rect(LineLeft, Buffer.Height - Border - iHiddenButtonHeight, LineRight, Buffer.Height - Border);
    if SelectedIndex = itHiddenArrow
    then begin
      if Options.Colors.Sunken <> clNone
      then DrawSolidBackground(Canvas, R, Options.Colors.Highlight)
      else DrawPatternBackground(Canvas, R);

      if (msLeftButton in MouseState)
      then Frame3D(Canvas, R, Options.Colors.FrameShadow, Options.Colors.Frame, 1)
      else Frame3D(Canvas, R, Options.Colors.Frame, Options.Colors.FrameShadow, 1);
    end
    else begin
      Brush.Style:= bsSolid;
      Brush.Color:= Options.Colors.Menu;
      FillRect(R);
    end;

    // draw arrow
    X:= (R.Right + R.Left) div 2 -3;
    Y:= (R.Top + R.Bottom) div 2 -4;
    if (SelectedIndex = itHiddenArrow)
    and (msLeftButton in MouseState)
    then begin
      Inc(X); Inc(Y);
    end;

    Pen.Color:= Options.Colors.MenuText;
    PolyLine([Point(X, Y), Point(X +2, Y +2), Point(X +4, Y), Point(X +4, Y +1), Point(X +2, Y +3),
      Point(X -1, Y)]);

    Inc(Y, 4);
    PolyLine([Point(X, Y), Point(X +2, Y +2), Point(X +4, Y), Point(X +4, Y +1), Point(X +2, Y +3),
      Point(X -1, Y)]);
  end;
end;

procedure TPopupMenu2000Form.PaintScrollBars;
var
  R: TRect;
  X, Y: Integer;
begin
  with Buffer, Canvas, PaintParams.mir
  do begin
    // draw bottom arrow
    if LastVisibleIndex < ItemRects.Count -1
    then begin
      R:= Rect(LineLeft, Buffer.Height - Border - iScrollButtonHeight -1, LineRight,
        Buffer.Height - Border);

      if SelectedIndex = itBottomScroll
      then begin
        DrawPatternBackground(Canvas, R);
        if msLeftButton in MouseState
        then DrawEdge(Canvas.Handle, R, bdr_SunkenOuter, bf_Rect)
        else DrawEdge(Canvas.Handle, R, bdr_RaisedInner, bf_Rect);
      end
      else begin
        Brush.Style:= bsSolid;
        Brush.Color:= Options.Colors.Menu;
        FillRect(R);
      end;

      // draw arrow
      X:= (R.Right + R.Left) div 2 -4;
      Y:= (R.Top + R.Bottom) div 2 -2;
      if (msLeftButton in MouseState)
      and (SelectedIndex = itBottomScroll)
      then begin
        Inc(X); Inc(Y);
      end;

      Pen.Color:= Options.Colors.MenuText;
      Brush.Style:= bsSolid;
      Brush.Color:= Options.Colors.MenuText;
      PolyGon([Point(X, Y), Point(X +3, Y +3), Point(X +6, Y)]);
    end;

    // draw top arrow
    if FirstVisibleIndex > 0
    then begin
      R:= Rect(LineLeft, Border, LineRight, Border + iScrollButtonHeight +1);

      if SelectedIndex = itTopScroll
      then begin
        DrawPatternBackground(Canvas, R);
        if msLeftButton in MouseState
        then DrawEdge(Canvas.Handle, R, bdr_SunkenOuter, bf_Rect)
        else DrawEdge(Canvas.Handle, R, bdr_RaisedInner, bf_Rect);
      end
      else begin
        Brush.Style:= bsSolid;
        Brush.Color:= Options.Colors.Menu;
        FillRect(R);
      end;

      // draw arrow
      X:= (R.Right + R.Left) div 2 -4;
      Y:= (R.Top + R.Bottom) div 2;
      if (msLeftButton in MouseState)
      and (SelectedIndex = itTopScroll)
      then begin
        Inc(X); Inc(Y);
      end;

      Pen.Color:= Options.Colors.MenuText;
      Brush.Style:= bsSolid;
      Brush.Color:= Options.Colors.MenuText;
      PolyGon([Point(X, Y), Point(X +3, Y -3), Point(X +6, Y)]);
    end;
  end;
end;

procedure TPopupMenu2000Form.Paint;
var
  X, Y, I, Temp: Integer;
  OldPalette: HPalette;
  M, M1, M0, M2: TMenuItem;
begin
  if (fsAnimated in State)
  and (Animation = anFadeIn)
  and (not (fsPaintMenu in State))
  then Exit;

  // init PaintParams record
  InitializePaintParams;

  Y:= PaintParams.mir.Border;
  if (Options.Draggable)
  and (BorderStyle = bsNone)
  then Inc(Y, 9);

  if BorderStyle <> bsNone
  then Inc(Y, 1);

  // select background's palette (if any)
  OldPalette:= 0;
  if Options.Background.Empty
  then
    Back.Handle:= 0
  else begin
    if (Options.Background.Palette <> 0)
    then begin
      OldPalette:= SelectPalette(Buffer.Canvas.Handle, Options.Background.Palette, True);
      RealizePalette(Buffer.Canvas.Handle);
    end;

    Include(PaintParams.State, isGraphBack);

    // fit background to menu
    if fsPaintMenu in State
    then PaintBackground;
  end;


  // draw frames
  if fsPaintMenu in State
  then begin
    PaintBevelAndBorder;
    PaintFrame3D;
  end;

  // mouse position
  GetCursorPos(PaintParams.MousePos);
  PaintParams.MousePos:= ScreenToClient(PaintParams.MousePos);

  // paint CanTearOff plane
  if (Options.Draggable)
  and ((fsPaintMenu in State)
    or ((fsSelectedChanged in State) and ((SelectedIndex = itDragPane)
      or (LastSelectedIndex = itDragPane))))
  and (BorderStyle = bsNone)
  then PaintDragPane;

  M:= nil;
  M1:= GetMenuItemIndex(FirstVisibleIndex);

  // top scroll
  if FirstVisibleIndex > 0
  then Inc(Y, iScrollButtonHeight);

  // repeat for each menu item
  for I:= FirstVisibleIndex to LastVisibleIndex
  do begin
    M0:= M;
    M:= M1;
    M1:= GetMenuItemIndex(I +1);
    Temp:= GetMenuItemHeightIndex(I);
    PaintParams.mir.Top:= Y;
    PaintParams.mir.Height:= Temp;

    // init state
    if (fsFraming in State)
    then begin
      SetState(PaintParams.State, isSelected, False);
      SetState(PaintParams.State, isFramed, (ActiveMenuItem = M));

      SetState(PaintParams.State, isInsertBefore, (ActiveMenuItem <> M)
        and (InsertBeforeItem = M));
      SetState(PaintParams.State, isInsertAfter, (ActiveMenuItem <> M)
        and (((M1 <> nil) and (InsertBeforeItem = M1))
        or ((M1 = nil) and (InsertBeforeItem = miInsertAtTheEnd))));
    end
    else
      SetState(PaintParams.State, isSelected, SelectedIndex = I);

    Exclude(PaintParams.State, isNoLeftSunken);
    Exclude(PaintParams.State, isNoRightSunken);
    if fsShowHidden in State
    then begin
      if not (fsHiddenAsRegular in State)
      then begin
        SetState(PaintParams.State, isHidden, (M is TMenuItem2000) and (TMenuItem2000(M).Hidden));
        SetState(PaintParams.State, isHiddenPrev, ((M0 is TMenuItem2000) and (TMenuItem2000(M0).Hidden))
          or ((M0 = nil) and (isHidden in PaintParams.State) and not Options.Draggable));
        SetState(PaintParams.State, isHiddenSucc, ((M1 is TMenuItem2000) and (TMenuItem2000(M1).Hidden))
          or ((M1 = nil) and (isHidden in PaintParams.State)));


        if (isHidden in PaintParams.State)
        then begin
          // TASK A: not-hidden item, hidden separator, hidden item -> unhide the separator
          // A1: remove isHiddenSucc at not-hidden item
          // excessive

          // A2: unhide separator itself
          if (not (isHiddenPrev in PaintParams.State))
          and (M <> nil)
          and (M.Caption <> '')
          and (M.Caption[1] = SSeparator)
          then Exclude(PaintParams.State, isHidden);

          // A3: remove isHiddenPrev at hidden menu item
          if (isHiddenPrev in PaintParams.State)
          and (M0 <> nil)
          and (M0.Caption <> '')
          and (M0.Caption[1] = SSeparator)
          then Exclude(PaintParams.State, isHiddenPrev);

          // TASK B: hidden item, hidden separator, not-hidden item -> unhide the separator
          // B1: remove isHiddenSucc at hidden menu item
          if (isHiddenSucc in PaintParams.State)
          and (M1 <> nil)
          and (M1.Caption <> '')
          and (M1.Caption[1] = SSeparator)
          then begin
            M2:= GetMenuItemIndex(I +2);
            if (M2 is TMenuItem2000)
            and (not TMenuItem2000(M2).Hidden)
            then Exclude(PaintParams.State, isHiddenSucc);
          end;

          // B2: unhide separator itself
          if (isHiddenPrev in PaintParams.State)
          and (not (isHiddenSucc in PaintParams.State))
          and (M <> nil)
          and (M.Caption <> '')
          and (M.Caption[1] = SSeparator)
          then Exclude(PaintParams.State, isHidden);

          // B3: remove isHiddenPrev at not-hidden menu item
          // excessive


          // TASK C&D: hise separator between two hidden menu items
          // hide next separator
          if (not (isHiddenSucc in PaintParams.State))
          and (M1 <> nil)
          and (M1.Caption <> '')
          and (M1.Caption[1] = SSeparator)
          then begin
            M2:= GetMenuItemIndex(I +2);
            if (M2 is TMenuItem2000)
            and (TMenuItem2000(M2).Hidden)
            then SetState(PaintParams.State, isHiddenSucc, True);
          end;

          // hide prev separator
          if (not (isHiddenPrev in PaintParams.State))
          and (M0 <> nil)
          and (M0.Caption <> '')
          and (M0.Caption[1] = SSeparator)
          then begin
            M2:= GetMenuItemIndex(I -2);
            if (M2 is TMenuItem2000)
            and (TMenuItem2000(M2).Hidden)
            then SetState(PaintParams.State, isHiddenPrev, True);
          end;

          if Options.Title.Visible
          then
            if Options.Title.Align = atLeft
            then Include(PaintParams.State, isNoLeftSunken)
            else Include(PaintParams.State, isNoRightSunken);

        end;
      end;
    end

    // ignore separator in Window menu
    else
      if (CurHiddenCount > 0)
      and (CurHiddenCount = MenuItems.Count)
      then Inc(CurHiddenCount);

    // handle focus item
    SetState(PaintParams.State, isActivated, (msLeftButton in MouseState) or ((FocusItem = M)
      and (fsDisabled in State)));


    PaintParams.Item:= M;
    PaintParams.Index:= I;

    // is menu item should be drawn?
    if (fsPaintMenu in State)
    or (fsFraming in State)
    or (((I = SelectedIndex) or (I = LastSelectedIndex))) then begin
      // draw background
      if not Back.Empty
      then
        with PaintParams.mir do
          BitBlt(Buffer.Canvas.Handle, Ctl3dWidth + LeftTitleWidth, Top,
            Width - 2*Ctl3dWidth - TitleWidth, Height,
            Back.Canvas.Handle, 0, Y + Ctl3dWidth, SrcCopy);

      // draw menu item
      // win32 menu item
      if M = nil
      then begin
        Inc(PaintParams.Index, CurHiddenCount);
        PaintMenuItemWin32(@PaintParams);
      end
      else
      // control
      if (M is TMenuItem2000) and (TMenuItem2000(M).Control <> ctlNone)
      then
        TMenuItem2000(M).ControlOptions.Paint(@PaintParams)

      else
      // ordinary menu item
        PaintMenuItem(@PaintParams);

    end;

    // increment Y
    Inc(Y, Temp);

    // fill interitem space
    if (fsPaintMenu in State)
    and (Options.Margins.Space > 0)
    and (I < ItemRects.Count -1)
    then
      if not Back.Empty
      then
        with PaintParams.mir do
          BitBlt(Buffer.Canvas.Handle, LineLeft, Y, LineRight - LineLeft,
            Options.Margins.Space, Back.Canvas.Handle, 0, Y, Buffer.Canvas.CopyMode)
      else
        with Buffer.Canvas, PaintParams.mir
        do begin
          if Brush.Style <> bsSolid
          then Brush.Style:= bsSolid;

          if Brush.Color <> Options.Colors.Menu
          then Brush.Color:= Options.Colors.Menu;

          Buffer.Canvas.FillRect(Rect(LineLeft, Y, LineRight,
            Y + Options.Margins.Space));
        end;

    // increment Y
    Inc(Y, Options.Margins.Space);
  end;

  Dec(Y, Options.Margins.Space);

  // final paintings
  if fsPaintMenu in State
  then PaintBackgroundPart(Y);

  // draw hidden arrow
  if (fsHiddenArrow in State)
  and ((fsPaintMenu in State)
    or (((fsSelectedChanged in State) or (fsMouseChanged in State))
      and ((SelectedIndex = itHiddenArrow) or (LastSelectedIndex = itHiddenArrow))
      and (not (fsFraming in State))))
  then PaintHiddenArrow;

  // draw scroll bars
  if (fsScrollbars in State)
  and ((fsPaintMenu in State)
    or (fsMouseChanged in State)
    or ((fsSelectedChanged in State)
      and ((SelectedIndex = itTopScroll)
      or (LastSelectedIndex = itTopScroll)
      or (SelectedIndex = itBottomScroll)
      or (LastSelectedIndex = itBottomScroll)
      or (SelectedIndex = LastVisibleIndex)
      or (LastSelectedIndex = LastVisibleIndex)
      )))
  then PaintScrollBars;

  // restore palette
  if OldPalette <> 0
  then SelectPalette(Buffer.Canvas.Handle, OldPalette, True);


  // determine buffer's position & draw the buffer
  if fsFromBottomToTop in State then Y:= 0 else Y:= Buffer.Height - Height;
  if fsFromRightToLeft in State then X:= 0 else X:= Buffer.Width - Width;

  if (BorderStyle <> bsNone)
  then Inc(Y, ClientOrigin.Y - Top - GetSystemMetrics(SM_CYCAPTION));

  BitBlt(Canvas.Handle, 0, 0, Buffer.Width, Buffer.Height, Buffer.Canvas.Handle, X, Y, SrcCopy);

  Exclude(State, fsPaintMenu);
  Exclude(State, fsMouseChanged);
  LastSelectedIndex:= SelectedIndex;
end;

procedure TPopupMenu2000Form.PaintCachedDesktop(DC: HDC; BufferX, BufferY: Integer);
  // paints cached desktop picture on child menu's buffer
begin
  if bits0 = nil then Exit;

  SetDIBits(Buffer.Canvas.Handle, Buffer.Handle, 0, MY, bits0, bi, Dib_Rgb_Colors);
  BitBlt(DC, Left - BufferX, Top - BufferY, Buffer.Width, Buffer.Height, Buffer.Canvas.Handle, 0, 0,
    SrcCopy);

  if ParentMenuForm <> nil
  then ParentMenuForm.PaintCachedDesktop(Buffer.Canvas.Handle, Left, Top);
end;


procedure TPopupMenu2000Form.RebuildToolTipWindow(Recreate: Boolean);
  // set tooltips
var
  R: TRect;
  M: TMenuItem;
  X, Y, H, I, iI, iJ, Index, BW, BH: Integer;
  S: String;
begin
  // we have to recreate tooltip window if it became
  // floating
  ToolTipWindow.Free;
  ToolTipWindow:= nil;

  if (not PopupMenu.ShowHint)
  or (PopupMenu.StatusBar <> nil)
  then Exit;

  ToolTipWindow:= T_AM2000_ToolTipWindow.Create(Self);

  // init X
  X:= 0;
  if Options.Title.Visible
  and (Options.Title.Align = atLeft)
  then Inc(X, Options.Title.Width);

  // init Y
  Y:= Options.Margins.Border;
  if (BorderStyle = bsNone)
  then
    if (fsCtl3D in State)
    then begin Inc(X, Options.Margins.Bevel +2); Inc(Y, Options.Margins.Bevel +2); end
    else begin Inc(X, Options.Margins.Bevel); Inc(Y, Options.Margins.Bevel); end;

  // is menu draggable?
  if (Options.Draggable)
  and (BorderStyle = bsNone)
  then begin
    ToolTipWindow.AddTool(Rect(2, Y, MX -2, Y+9), SDraggableMenuInfo);
    Inc(Y, 10);
  end;


  // add tools to the ToolTip control
  for I:= 0 to ItemRects.Count -1 do begin
    M:= GetMenuItemIndex(I);
    H:= GetMenuItemHeightIndex(I);

    if (M <> nil) then
      // is it a button array?
      if (M is TMenuItem2000)
      and (TMenuItem2000(M).Control = ctlButtonArray)
      then
        with TMenuItem2000(M).AsButtonArray do begin
          if Bitmap.Empty then Continue;

          // init loop
          iI:= 0;
          iJ:= 0;
          Index:= 0;
          BW:= Bitmap.Width div Columns;
          BH:= Bitmap.Height div Rows;

          // assign each hint to tooltip control
          while Index < Hints.Count do begin
            R:= Rect(iI * BW + X +2, iJ * BH + Y +1, 0, 0);
            R.Right:= R.Left + BW;
            R.Bottom:= R.Top + BH;

            if Hints[Index] <> ''
            then ToolTipWindow.AddTool(R, Hints[Index]);

            // next step
            Inc(Index);
            Inc(iI);
            if iI = Columns
            then begin Inc(iJ); iI:= 0; end;
          end;
        end

      // it's usual item
      else begin
        // is hint present?
        S:= '';
        if (M.Hint <> '') and (M.Hint <> #1)
        then S:= M.Hint;

        if (S = '')
        and (M.Hint = '')
        and (MenuItemCache.IsDefault(M.Caption))
        then S:= MenuItemCache[M.Caption].Hint;

        if S <> ''
        then ToolTipWindow.AddTool(Rect(0, Y, MX, Y+H), S);
      end;

    Inc(Y, H + Options.Margins.Space);
  end;

  // hidden arrow
  if (fsHiddenArrow in State)
  then ToolTipWindow.AddTool(Rect(0, Y, MX, MY), SExpandCaption);

  ToolTipWindow.Activate;
end;


procedure TPopupMenu2000Form.RebuildBounds;
var
  I, DI1, DI2, H, P, TW, TH, W1, W2, MIC, W11, P11: Integer;
  M: TMenuItem;
  L: LongRec;
  S, LastCaption: String;
  IncludeBold: Boolean;

  procedure SetDefault(D: Boolean);
  begin
    with Buffer.Canvas.Font do
      if D then begin
        if not (fsBold in Style) then begin
          IncludeBold:= True;
          Style:= Style + [fsBold];
        end;
      end
      else
        if IncludeBold then begin
          IncludeBold:= False;
          Style:= Style - [fsBold];
        end;
  end;

begin
  // initialize
  P:= 0;
  M:= nil;
  I:= 0;
  DI1:= 0; // for invisible items
  DI2:= 0; // for hidden items
  ItemWidth:= 0;
  ShortcutWidth:= 0;
  IncludeBold:= False;
  FiWidth:= 0;
  Exclude(State, fsHiddenArrow);
  ItemRects.Clear;

  if (MenuHandle = 0) or (Options = nil)
  then Exit;

  MIC:= GetMenuItemCount(MenuHandle);

  // start
  // calc top margin
  FiLeft:= Options.Margins.Border + Options.Margins.Bevel;
  H:= Options.Margins.Border + Options.Margins.Bevel;
  if (BorderStyle = bsNone)
  then begin
    if (fsCtl3D in State)
    then begin Inc(H, 2); Inc(FiLeft, 2); end;

    // is menu draggable?
    if (Options.Draggable)
    then Inc(H, iDraggablePaneHeight +2);
  end;
//  else
//    Inc(H, 1);

  // calc all menu items
  while MIC > I + DI2
  do begin
    W1:= 0;
    W2:= 0;
    L.Lo:= H;
    L.Hi:= 0;

    // calc item's heigth
    // Delphi menu item
    while (MenuItems <> nil)
    and (MenuItems.Count > I + DI1 + DI2)
    do begin
      M:= MenuItems[I + DI1 + DI2];

      // invisible item
      if (not M.Visible)
      then
        Inc(DI1)
      else
      // hidden item
      if (M is TMenuItem2000)
      and (TMenuItem2000(M).Hidden)
      and (not (fsShowHidden in State))
      then
        Inc(DI2)
      else
      // visible item
      begin
        L.Hi:= GetMenuItemHeight(M);

        // create new font style
        SetDefault(M.Default);
        S:= M.Caption;
        LastCaption:= M.Caption;
        Break;
      end;
    end;

    // WinAPI menu item
    if M = nil
    then begin
      if (MenuItems = nil)
      or (MIC > MenuItems.Count)
      then begin
        mii.fMask:= $3F or miim_Type or miim_Submenu or miim_State;
        mii.dwTypeData:= Z;
        mii.cch:= SizeOf(Z) -1;
        if GetMenuItemInfo(MenuHandle, I + DI2, True, mii)
        then
          if (I = 0)
          and (DI2 <> 0)
          and (mii.fType and mft_Separator <> 0)
          then begin // ignore first separator in the menu
            Inc(I);
            Continue;
          end
          else begin // not the first separator in menu
            if mii.fType and mft_Separator <> 0
            then L.Hi:= Options.SeparatorHeight
            else L.Hi:= ItemHeight;

            SetDefault(mii.fState and mfs_Default <> 0);
            S:= StrPas(Z);
          end;
      end

      else
        Break;
    end;
    // end empty M

    // calc item's caption width
    if (M <> nil)
    and (M is TMenuItem2000)
    then begin
      W1:= TMenuItem2000(M).GetWidth(Buffer.Canvas, Options);
      if W1 > 0 then Dec(W1, Options.Margins.Left + Options.Margins.Right);
    end;

    if W1 = 0
    then begin
      P:= Pos(#9, S);
      if P = 0 then P:= Pos(STabulator, S);
      if P <> 0 then S:= Copy(S, 1, P -1);

      // calc menu item's width
      repeat
        P11:= Pos(SNewLine, S);
        if P11 = 0 then P11:= Pos(#13, S);
        if P11 = 0 then P11:= Length(S) +1;

        if (S = '') or (P11 = 0) then Break;

        W11:= AmpTextWidth(Buffer.Canvas, Copy(S, 1, P11 -1));
        if W11 > W1 then W1:= W11;

        if (P11 < Length(S))
        and (S[P11] = #13)
        then Delete(S, 1, P11)
        else Delete(S, 1, P11 +1);
      until S = '';
    end;

    // add triangle
//    if ((M <> nil) and HasSubmenu(M))
//    or (mii.hSubmenu <> 0)
//    then begin
//      Inc(W1, iSubmenuTriangleWidth);
//    end;

    if W1 > iMaxItemWidth then W1:= iMaxItemWidth;
    // end caption width

    // calc item's shortcut width
    if (M <> nil)
    then begin
      if (M is TMenuItem2000)
      then begin
        if TMenuItem2000(M).ShortCut <> ''
        then W2:= Buffer.Canvas.TextWidth(GetMainShortCut(TMenuItem2000(M).ShortCut))
      end
      else
      if M.ShortCut <> 0
      then
        W2:= Buffer.Canvas.TextWidth(ShortCutToText(M.ShortCut));

      // custom shortcut
      if (W2 = 0)
      and (P <> 0)
      then begin
        S:= M.Caption;
        if S[P] = #9
        then S:= Copy(S, P +1, MaxInt)
        else S:= Copy(S, P +2, MaxInt);

        W2:= Buffer.Canvas.TextWidth(S);
      end;
    end
    else
      if P <> 0
      then W2:= Buffer.Canvas.TextWidth(Copy(Z, P +1, MaxInt));
    // end shortcut width


    // OnMeasureItem event handler
    // has OnMeasureItem event handler?
    if (M <> nil)
{$IFNDEF Delphi4OrHigher}
    and (M is TMenuItem2000)
{$ENDIF}
    and Assigned(TMenuItem2000(M).OnMeasureItem)
    then begin
      TW:= W1 + W2;
      TH:= L.Hi;
      TMenuItem2000(M).OnMeasureItem(M, Buffer.Canvas, TW, TH);

      if TW <> W1 + W2 then W1:= TW -W2;
      L.Hi:= TH;

    end;

    // add item's top and botton
    if L.Hi <> 0 then begin
      ItemRects.Add(Pointer(L));
      Inc(H, L.Hi + Options.Margins.Space);
      Inc(I);
    end;


    // max shortcut
    if ShortcutWidth < W2 then ShortcutWidth:= W2;

    // merge max caption width and max shrtcut width
    // office-like align (align shortcuts to the left)
    if not (mfStandardAlign in Options.Flags) then begin
      if W2 > 0 then Inc(W2, 4);
      if (FiWidth < W1 + W2) then FiWidth:= W1 + W2;
    end;

    // max width
    if ItemWidth < W1 then ItemWidth:= W1;

    M:= nil;
  end { loop for each menu item };

  // auto line reduction
  if (LastCaption <> '')
  and (LastCaption[1] = SSeparator)
  then begin
    ItemRects.Delete(ItemRects.Count -1);
  end;

  // restore default canvas settings
  SetDefault(False);

  // merge max caption width and max shortcut width
  // standard align (align shortcuts to the right)
  if (mfStandardAlign in Options.Flags) then begin
    FiWidth:= ItemWidth + ShortCutWidth;
    if ShortCutWidth > 0 then Inc(FiWidth, 4);
  end;

  // add margins
  Inc(FiWidth, Options.Margins.Left + Options.Margins.Right);

  // add title
  with Options.Title
  do
    if (ParentMenuForm = nil)
    and Visible
    and (Align = atLeft)
    then Inc(FiLeft, Width);

  // set flag for hidden menu items
  if DI2 > 0
  then Include(State, fsHiddenArrow);

  // check menu with hidden menu items
  // and without unhidden menu items
  if (I = 0) and (DI2 > 0)
  then begin
    Include(State, fsShowHidden);
    RebuildBounds;
  end;
end;

function TPopupMenu2000Form.GetItemRect(Index: Integer): TRect;
begin
  if (Index >= 0) and (Index < ItemRects.Count)
  then begin
    Result.Left:=   FiLeft;
    Result.Top:=    LongRec(ItemRects[Index]).Lo;
    Result.Right:=  Result.Left + FiWidth;
    Result.Bottom:= Result.Top + LongRec(ItemRects[Index]).Hi;
  end
  else
    FillChar(Result, SizeOf(TRect), 0);
end;

procedure TPopupMenu2000Form.wmInitState;
var
  HiddenIsVisible: Boolean;
begin
  // init menu items
  if Owner is TForm
  then
    SendMessage(TForm(Owner).Handle, wm_InitMenuPopup, LongInt(MenuHandle), 0);

  if (PopupMenu.Ole2ObjectWindow <> 0)
  and not (PopupMenu.MenuPopulated)
  then
    SendMessage(PopupMenu.Ole2ObjectWindow, wm_InitMenuPopup, LongInt(MenuHandle), 0);


  FormStyle:= fsStayOnTop;
  BorderStyle:= bsNone;

  HiddenIsVisible:= (fsShowHidden in State);
  if (PopupMenu = nil) then Exit;

  if HiddenIsVisible
  then
    PopupMenu.Options.Flags:= PopupMenu.Options.Flags + [mfHiddenIsVisible];

  // set Ctl3D
  if (PopupMenu.Ctl3D)
  then State:= [fsCtl3d]
  else State:= [];

  // assign options property
  GetOptions(MenuItems, PopupMenu, Options);

  Color:= Options.Colors.Menu;

  // for submenus these properties
  // already initialized in CreateSubMenuForm
  if ParentMenuForm = nil
  then
    Animation:= Options.Animation;

  // hidden
  if HiddenIsVisible
  or (mfHiddenIsVisible in Options.Flags)
  then
    Include(State, fsShowHidden);

  if (mfHiddenAsRegular in Options.Flags)
  then
    Include(State, fsHiddenAsRegular);

  if PopupMenu.ShowAccelerators
  then
    Include(State, fsShowAccelerators);

  // reassign font handle
  if PopupMenu.ParentFont
  then begin
    if (PopupMenu.Owner is TControl)
    then Font:= TForm(PopupMenu.Owner).Font
    else
    if (PopupMenu.PopupComponent <> nil)
    and (PopupMenu.PopupComponent is TControl)
    then Font:= TForm(PopupMenu.PopupComponent.Owner).Font
  end
  else
    if PopupMenu.SystemFont
    then Font.Handle:= GetMenuFontHandle
    else Font:= PopupMenu.Font;

  Canvas.Font.Assign(Font);
  Buffer.Canvas.Font.Assign(Font);

  // get menu item height
  if (Options.ItemHeight = 0)
  then
    ItemHeight:= Canvas.TextHeight('Hj') +4
  else
    ItemHeight:= Options.ItemHeight;

  // reset Top and left
  if (ParentMenuForm = nil)
  and (PopupMenu <> nil)
  then SetBounds(PopupMenu.PopupX, PopupMenu.PopupY, 0, 0);

  // fsFraming
  if (PopupMenu <> nil)
  and (PopupMenu.PopupComponent is TCustomMenuBar2000)
  and (TCustomMenuBar2000(PopupMenu.PopupComponent).aiState = aiFraming)
  then begin
    Include(State, fsFraming);
    SetDragCursor;
  end;
end;

procedure TPopupMenu2000Form.wmSetDesigning(var Msg: TMessage);
begin
  if Msg.wParam <> 0
  then begin
    Include(State, fsDesigning);
  end
  else begin
    Exclude(State, fsDesigning);
  end;

  Msg.Result:= 1;
end;

procedure TPopupMenu2000Form.OnHotTrack(Sender: TObject);
begin
  StopTimer;
  KillActivePopupMenu2000(True, True);
end;

{$IFDEF Delphi4OrHigher}
procedure TPopupMenu2000Form.DoClose(var Action: TCloseAction);
begin
  inherited;
  Action:= caFree;
end;
{$ENDIF}

procedure TPopupMenu2000Form.Refresh;
var
  NewLeft, NewTop, NewHeight: Integer;
begin
  RebuildBounds;
  RebuildToolTipWindow(False);

  MX:= GetRealWidth;
  MY:= GetRealHeight;

  CheckOrigin(anNone, NewLeft, NewTop, NewHeight);
  SetBounds(NewLeft, NewTop, MX, MY);

  Buffer.Width:= MX;
  Buffer.Height:= MY;
  Buffer.FreeImage;
  Include(State, fsPaintMenu);
  Paint;

  if SubMenuForm <> nil
  then SubMenuForm.Refresh;
end;

procedure TPopupMenu2000Form.wmSetEnabled(var Msg: TMessage);
begin
  if Msg.wParam <> 0
  then Exclude(State, fsDisabled)
  else Include(State, fsDisabled);
end;

procedure TPopupMenu2000Form.wmSetFocusItem(var Msg: TMessage);
begin
  FocusItem:= TMenuItem2000(Msg.wParam);
end;

end.
